"""
@file convergence.py
Convergence python backend.
"""
from hysop.constants import HYSOP_REAL
from hysop.backend.host.host_operator import HostOperator
from hysop.operator.base.convergence import ConvergenceBase
from hysop.tools.decorators  import debug
from hysop.core.graph.graph import op_apply
from hysop.tools.numpywrappers import npw
from hysop.constants import ResidualError
import mpi4py.MPI as MPI
import numpy as np


class PythonConvergence(ConvergenceBase, HostOperator):
    """Computes convergence of a field through iterations"""

    @debug
    def __init__(self, **kwds):
        super(PythonConvergence, self).__init__(**kwds)

    @debug
    def setup(self, **kwds):
        super(PythonConvergence, self).setup(**kwds)
        self.field_buffers = self.dField.compute_buffers

        self._tmp_convergence = npw.zeros((self.field.nb_components),
                                          dtype=self.convergence.dtype)
        self._tmp_reduce = npw.zeros((self.field.nb_components),
                                     dtype=self.convergence.dtype)
        old = [npw.zeros(_.shape) for _ in self.field_buffers]
        self.dField_old = tuple(old)

        if self._residual_computation == ResidualError.ABSOLUTE:
            self.__compute_error = lambda ui, ui_old, max_ui: npw.max(npw.abs(ui - ui_old))
        elif self._residual_computation == ResidualError.RELATIVE:
            self.__compute_error = lambda ui, ui_old, max_ui: npw.max(npw.abs(ui - ui_old))/max_ui
        else:
            raise RuntimeError('Unknown residual computation method.')

        self._eps = npw.finfo(HYSOP_REAL).eps
        self._large_zero = 1e3 * npw.finfo(HYSOP_REAL).eps

    @op_apply
    def apply(self, **kwds):
        u   = self.field_buffers
        u_old = self.dField_old

        self._tmp_convergence[...] = 0.
        for (i, (ui, ui_old)) in enumerate(zip(u, u_old)):
            max_ui = npw.max(npw.abs(ui))
            if max_ui < self._large_zero:
                self._tmp_convergence[i] = self._eps
            else:
                self._tmp_convergence[i] = self.__compute_error(ui, ui_old, max_ui)
            ui_old[...] = ui
        self.mpi_params.comm.Allreduce(sendbuf=self._tmp_convergence,
                                       recvbuf=self._tmp_reduce,
                                       op=MPI.MAX)
        self.convergence.value = self._tmp_reduce


    @classmethod
    def supports_mpi(cls):
        return True
