from hysop.backend.host.host_operator import HostOperator
from hysop.tools.types import check_instance, InstanceOf
from hysop.tools.decorators import debug
from hysop.fields.continuous_field import Field
from hysop.parameters.scalar_parameter import ScalarParameter
from hysop.topology.cartesian_descriptor import CartesianTopologyDescriptors
from hysop.methods import SpaceDiscretization
from hysop.core.memory.memory_request import MemoryRequest
from hysop.core.graph.graph import op_apply
from hysop.numerics.stencil.stencil_generator import StencilGenerator, CenteredStencilGenerator, MPQ
import numpy as np


class PythonPenalizeVorticity(HostOperator):
    """

    """
    __default_method = {
        SpaceDiscretization: SpaceDiscretization.FDC4
    }

    __available_methods = {
        SpaceDiscretization: (InstanceOf(SpaceDiscretization), InstanceOf(int))
    }

    @classmethod
    def default_method(cls):
        dm = super(PythonPenalizeVorticity, cls).default_method()
        dm.update(cls.__default_method)
        return dm
    @classmethod
    def available_methods(cls):
        am = super(PythonPenalizeVorticity, cls).available_methods()
        am.update(cls.__available_methods)
        return am

    @debug
    def __init__(self, obstacles, variables,
                 velocity, vorticity,
                 dt, coeff=None, **kwds):
        check_instance(velocity, Field)
        check_instance(vorticity, Field)
        check_instance(variables, dict, keys=Field,
                       values=CartesianTopologyDescriptors)
        check_instance(dt, ScalarParameter)
        check_instance(coeff, (ScalarParameter, float, type(lambda x:x)))
        check_instance(obstacles, (tuple, dict), values=Field,
                       keys=(ScalarParameter, float), check_kwds=False)

        input_fields = {velocity: variables[velocity],
                        vorticity: variables[vorticity]}
        output_fields = {vorticity: variables[vorticity]}
        input_params = {dt.name: dt}

        if isinstance(coeff, ScalarParameter):
            self.coeff = lambda o: coeff()*o
        if isinstance(coeff, type(lambda x:x)):
            self.coeff = coeff
        else:
            c = ScalarParameter("penal_coeff", initial_value=coeff)
            self.coeff = lambda o: c()*o

        if isinstance(obstacles, dict):
            obs = obstacles
        else:
            obs = {}
            for o in obstacles:
                obs[self.coeff] = o
        for o in obs.values():
            assert o.nb_components == 1
            input_fields[o] = variables[o]

        self.velocity = velocity
        self.vorticity = vorticity
        self.obstacles = obs
        self.dt = dt
        super(PythonPenalizeVorticity, self).__init__(
            input_fields=input_fields, output_fields=output_fields,
            input_params=input_params, **kwds)

    @debug
    def handle_method(self, method):
        super(PythonPenalizeVorticity, self).handle_method(method)
        self.space_discretization = method.pop(SpaceDiscretization)
        csg = CenteredStencilGenerator()
        csg.configure(dtype=MPQ, dim=1)
        try:
            self.order = int(str(self.space_discretization)[3:])
        except:
            self.order = self.space_discretization
        self.stencil = tuple(csg.generate_exact_stencil(
            derivative=1,
            order=self.order) for _ in range(3))

    @debug
    def get_field_requirements(self):
        requirements = super(PythonPenalizeVorticity, self).get_field_requirements()
        stencil = self.stencil[0]
        G = max(max(a, b) for a, b in zip(stencil.L, stencil.R))
        for is_input, (field, td, req) in requirements.iter_requirements():
            min_ghosts = (max(g, G) for g in req.min_ghosts.copy())
            req.min_ghosts = min_ghosts
            req.axes = ((0, 1, 2), )
        return requirements

    @debug
    def discretize(self):
        if self.discretized:
            return
        super(PythonPenalizeVorticity, self).discretize()
        self.dvelocity = self.input_discrete_tensor_fields[self.velocity]
        self.dvorticity = self.input_discrete_tensor_fields[self.vorticity]

        dv, dw = self.dvelocity, self.dvorticity
        stencil = self.stencil[0]
        G = max(max(a, b) for a, b in zip(stencil.L, stencil.R))
        view = dv.local_slices(ghosts=(G, G, G))
        V = tuple(Vi[view] for Vi in dv.buffers)
        view = dw.local_slices(ghosts=(G, G, G))
        W = tuple(Wi[view] for Wi in dw.buffers)
        self.W, self.V = W, V
        self.dobstacles = {}
        for c, o in self.obstacles.iteritems():
            o_df = self.input_discrete_fields[o]
            self.dobstacles[c] = o_df.data[0][o_df.local_slices(ghosts=(G, G, G))]

        for s, dx in zip(self.stencil, dw.space_step):
            s.replace_symbols({s.dx: dx})

        self.ghost_exchanger = dw.build_ghost_exchanger()

    @debug
    def get_work_properties(self):
        requests  = super(PythonPenalizeVorticity, self).get_work_properties()
        buffers = MemoryRequest.empty_like(
            a=self.V[0], nb_components=4, backend=self.dvelocity.backend)
        requests.push_mem_request('Wtmp', buffers)
        return requests

    def setup(self, work):
        super(PythonPenalizeVorticity, self).setup(work=work)
        Wtmp = work.get_buffer(self, 'Wtmp', handle=True)
        self.tmp_v = Wtmp[0:3]
        self.tmp = Wtmp[-1]

    @classmethod
    def supported_dimensions(cls):
        return (3, )

    @classmethod
    def supports_mpi(cls):
        return True

    @op_apply
    def apply(self, **kwds):
        super(PythonPenalizeVorticity, self).apply(**kwds)
        dt = self.dt()
        V, W = self.V, self.W

        # Penalize velocity
        self.tmp[...] = 0.
        for c, o in self.dobstacles.iteritems():
            self.tmp[...] += - dt * c(o) / (1.0 +  dt * c(o))
        for v, tmp in zip(V, self.tmp_v):
            tmp[...] = v * self.tmp

        # compute penalized vorticity:
        #     (dvz/dy - dvy/dz)
        # W = (dvx/dz - dvz/dx)
        #     (dvy/dx - dvx/dy)
        # X direction
        self.tmp = self.stencil[0](a=self.tmp_v[2], out=self.tmp, axis=2)
        W[1][...] += -self.tmp
        self.tmp = self.stencil[0](a=self.tmp_v[1], out=self.tmp, axis=2)
        W[2][...] += self.tmp

        # Y direction
        self.tmp = self.stencil[1](a=self.tmp_v[0], out=self.tmp, axis=1)
        W[2][...] += -self.tmp
        self.tmp = self.stencil[1](a=self.tmp_v[2], out=self.tmp, axis=1)
        W[0][...] += self.tmp

        # Z direction
        self.tmp = self.stencil[2](a=self.tmp_v[1], out=self.tmp, axis=0)
        W[0][...] += -self.tmp
        self.tmp = self.stencil[2](a=self.tmp_v[0], out=self.tmp, axis=0)
        W[1][...] += self.tmp

        self.ghost_exchanger()
