
import inspect
import graph_tool as gt
from graph_tool     import Graph, GraphView
from graph_tool     import topology, stats, search
from hysop.tools.decorators import not_implemented, debug, wraps, profile
    
class ComputationalGraphNodeData(object):
    """
    Simple class to hold some node data.
    """
    def __init__(self, current_level, node_id):
        self.current_level = current_level
        self.node_id       = node_id
        self.apply_kargs   = []  #list of dictionnary, last one has priority 
    def __str__(self):
        return '(lvl={},id={})'.format(self.current_level, self.node_id)

if __debug__:
    # python in debug mode, all decorators do check their target attribute

    def not_initialized(f):
        assert callable(f)
        @wraps(f)
        def _not_initialized(*args,**kargs):
            return f(*args,**kargs)
            self = args[0]
            msg = 'Cannot call {}.{}() on node \'{}\' because {}'\
                    .format(self.__class__.__name__,f.__name__,self.name,'{}')
            if self.initialized:
                reason='this node has already been initialized.'
                raise RuntimeError(msg.format(reason))
        return _not_initialized

    def initialized(f):
        assert callable(f)
        @wraps(f)
        def _initialized(*args,**kargs):
            self = args[0]
            msg = 'Cannot call {}.{}() on node \'{}\' because {}'\
                    .format(self.__class__.__name__,f.__name__,self.name,'{}')
            if not self.initialized:
                reason='this node has not been initialized yet.'
                raise RuntimeError(msg.format(reason))
            return f(*args,**kargs)
        return _initialized

    def discretized(f):
        assert callable(f)
        @wraps(f)
        def _discretized(*args,**kargs):
            self = args[0]
            msg = 'Cannot call {}.{}() on node \'{}\' because {}'\
                    .format(self.__class__.__name__,f.__name__,self.name,'{}')
            if not self.discretized:
                reason='this node has not been discretized yet.'
                raise RuntimeError(msg.format(reason))
            return f(*args,**kargs)
        return _discretized

    def ready(f):
        assert callable(f)
        @wraps(f)
        def _ready(*args,**kargs):
            self = args[0]
            msg = 'Cannot call {}.{}() on node \'{}\' because {}'\
                    .format(self.__class__.__name__,f.__name__, self.name,'{}')
            if not self.ready:
                reason='this node has not been set up.'
                raise RuntimeError(msg.format(reason))
            return f(*args,**kargs)
        return _ready

    def graph_built(f):
        assert callable(f)
        @wraps(f)
        def _graph_built(*args,**kargs):
            self = args[0]
            msg = 'Cannot call {}.{}() on node \'{}\' because {}'\
                    .format(self.__class__.__name__,f.__name__,self.name,'{}')
            if not self.graph_built:
                reason = 'the graph has not been built yet.'
                raise RuntimeError(msg.format(reason))
            return f(*args,**kargs)
        return _graph_built

    def generated(f):
        assert callable(f)
        @wraps(f)
        def _generated(*args,**kargs):
            self = args[0]
            msg = 'Cannot call {}.{}() on node \'{}\' because {}'\
                    .format(self.__class__.__name__,f.__name__,self.name,'{}')
            if not self.generated:
                reason='this node has not been generated yet.'
                raise RuntimeError(msg.format(reason))
            return f(*args,**kargs)
        return _generated

else: # not __debug__
    # python optimized, no checks
    def not_initialized(f):
        return f

    def initialized(f):
        return f

    def discretized(f):
        return f

    def ready(f):
        return f

    def graph_built(f):
        return f

    def generated(f):
        return f


def op_apply(f):
    @debug
    @profile
    @ready
    def apply(*args, **kwds):
        #print u'APPLY {}'.format(args[0].name)
        dbg = ('dbg' in kwds) 
        dbg = dbg and (kwds['dbg'] is not None)
        dbg = dbg and (kwds['dbg'].enable_on_op_apply)
        debug_dump = ('debug_dumper' in kwds) 
        debug_dump = debug_dump and (kwds['debug_dumper'] is not None)
        debug_dump = debug_dump and (kwds['debug_dumper'].enable_on_op_apply)
        if debug_dump:
            assert 'simulation' in kwds
            op = args[0]
            simu = kwds['simulation']
            it = simu.current_iteration
            t = simu.t()
            _file = inspect.getsourcefile(f)
            _, _line = inspect.getsourcelines(f)
            description='{}:{}'.format(_file, _line)
            for param in sorted(op.input_params.values(), key=lambda x: x.name):
                tag = 'pre_{}_{}'.format(op.name, param.name)
                kwds['debug_dumper'](it, t, tag, 
                        (param._value,), description=description)
            for dfield in sorted(op.input_discrete_fields.values(), key=lambda x: x.name):
                tag = 'pre_{}_{}'.format(op.name, dfield.name)
                kwds['debug_dumper'](it, t, tag, 
                        tuple(df.sdata.get().handle[df.compute_slices] for df in dfield.dfields), description=description)
            ret = f(*args, **kwds)
            for param in sorted(op.output_params.values(), key=lambda x: x.name):
                tag = 'post_{}_{}'.format(op.name, param.name)
                kwds['debug_dumper'](it, t, tag, 
                        (param._value,), description=description)
            for dfield in sorted(op.output_discrete_fields.values(), key=lambda x: x.name):
                tag = 'post_{}_{}'.format(op.name, dfield.name)
                kwds['debug_dumper'](it, t, tag, 
                        tuple(df.sdata.get().handle[df.compute_slices] for df in dfield.dfields), description=description)
            return ret
        elif dbg:
            msg=inspect.getsourcefile(f)
            kwds['dbg']('pre '+msg, nostack=True)
            ret = f(*args, **kwds)
            kwds['dbg']('post '+msg, nostack=True)
            return ret
        else:
            return f(*args, **kwds)
        return ret
    return apply
