"""Discrete advection for GPU
"""
from abc import abstractmethod, ABCMeta
from hysop import __VERBOSE__
from hysop.constants import np, debug, DirectionLabels, HYSOP_REAL
from hysop.methods import TimeIntegrator, Remesh, \
    Support, Splitting, MultiScale, Precision
from hysop.numerics.odesolvers import Euler
from hysop.operator.discrete.particle_advection import ParticleAdvection
from hysop.operator.discrete.discrete import get_extra_args_from_method
from hysop.backend.device.opencl import cl
from hysop.backend.device.opencl.opencl_kernel import OpenClKernelLauncher
import hysop.default_methods as default
from hysop.tools.numpywrappers import npw
from hysop.backend.device.opencl.opencl_discrete import OpenClDiscreteField
from hysop.backend.device.opencl.gpu_operator import GPUOperator
from hysop.tools.profiler import profile
from hysop.numerics.update_ghosts import UpdateGhostsFull
from hysop.tools.misc import WorkSpaceTools


class GPUParticleAdvection(ParticleAdvection, GPUOperator):
    """Particle advection solver on GPU

    """
    __metaclass__ = ABCMeta

    @debug
    def __init__(self, **kwds):
        """Particular advection of field(s) in a given direction,
        on GPU.

        OpenCL kernels are build once per dimension in order to handle
        directional splitting with resolution non uniform in directions.

        Parameters
        ----------
        kwds : base classes parameters

        Note
        ----
        * warning : this operator is derived from ParticleAdvection AND
        GPUOperator. kwds must then handle argument of both classes.
        """
        # Set default method if unknown
        if 'method' not in kwds:
            kwds['method'] = default.ADVECTION
            kwds['method'][Support] = 'gpu_2k'

        # init base class (i.e. ParticleAdvection)
        super(GPUParticleAdvection, self).__init__(**kwds)

        # mpi comm is set to advected field topology communicator.
        self.fields_topo = self.fields_on_grid[0].topology
        self._comm = self.fields_topo.comm
        self._comm_size = self._comm.Get_size()
        self._comm_rank = self._comm.Get_rank()
        assert self._comm == self.velocity.topology.comm

        # init second base class (i.e GPUOperator)
        if Precision in self.method:
            gpu_precision = self.method[Precision]
        else:
            gpu_precision = None
        GPUOperator.__init__(
            self, topology=self.velocity.topology,
            direction=self.direction,
            platform_id=get_extra_args_from_method(self, 'platform_id', None),
            device_id=get_extra_args_from_method(self, 'device_id', None),
            device_type=get_extra_args_from_method(self, 'device_type', None),
            gpu_precision=gpu_precision)

        # Choose between one kernel for all operations (default) ...
        self._use_2_kernels = False
        if self.method[Support].find('gpu_2k') >= 0:
            # ... or two different kernels for advection and remesh
            self._use_2_kernels = True

        self._synchronize = None
        self._isMultiScale = False
        if MultiScale in self.method and self.method[MultiScale] is not None:
            self._isMultiScale = True
            self._synchronize = UpdateGhostsFull(
                self.velocity.topology, self.velocity.nb_components)

        # Compute resolutions for kernels for each direction.
        # Resolution of the local mesh but reoganized regarding
        # splitting direction:
        # direction X : XYZ
        # direction Y : YXZ
        # direction Z : ZYX in parallel, ZXY in sequentiel.
        self.resol_dir = npw.dim_ones((self._dim,))
        self.v_resol_dir = npw.dim_ones((self._dim,))
        shape = self.fields_topo.mesh.local_resolution
        v_shape = self.velocity.topology.mesh.local_resolution
        # Local mesh resolution
        self.resol_dir[:self._dim] = self._reorderVect(shape)
        self.v_resol_dir[:self._dim] = self._reorderVect(v_shape)
        self._init_size_constants()

        # Init mesh info
        self._cl_mesh_info = None
        self._init_cl_mesh_info()

        # user defined opencl sources
        self.prg = None
        # get extra (opencl kernels) files from method, if any
        self._collect_usr_cl_src()

        # Set copy kernel
        self.copy = self._collect_kernels_cl_src_copy()

        # Set transposition kernels
        self.transpose_xy, self.transpose_xy_r = None, None
        self.transpose_xz, self.transpose_xz_r = None, None
        self._collect_kernels_cl_src_transpositions_xy()
        if self._dim == 3:
            self._collect_kernels_cl_src_transpositions_xz()

        # Set advection and remesh kernels
        self.num_advec, self.num_remesh = None, None
        self.num_advec_and_remesh = None
        if self._use_2_kernels:
            self._collect_kernels_cl_src_2k()
            self._compute = self._compute_2k
        else:
            self._collect_kernels_cl_src_1k()
            if self._isMultiScale:
                self._compute = self._compute_1k_multiechelle
            else:
                if self.method[TimeIntegrator] is Euler:
                    self._compute = self._compute_1k_euler_simpleechelle
                else:
                    self._compute = self._compute_1k_simpleechelle

        self._buffer_allocations()
        if self.direction == 0:
            self._buffer_initialisations()

        # List of executions
        self.exec_list = None
        self._build_exec_list()

        # Particle initialisation OpenCL events for each field:
        self._init_events = {self.fields_on_grid[0]: []}

    def _init_size_constants(self):
        """Fill size_constants attribute in
        (_size_constant belongs to gpu_operator)
        """
        v_shape = self.velocity.topology.mesh.local_resolution
        f_shape = self.fields_topo.mesh.local_resolution
        self._append_size_constants(f_shape)
        self._append_size_constants(v_shape, prefix='V_NB')
        self._append_size_constants(
            [self.velocity.topology.ghosts()[self.direction]],
            prefix='V_GHOSTS_NB', suffix=[''])
        enum = ['I', 'II', 'III']
        self._append_size_constants(
            self._reorderVect(['NB' + d for d in DirectionLabels[:self._dim]]),
            prefix='NB_', suffix=enum[:self._dim])
        self._append_size_constants(
            self._reorderVect(['V_NB' + d for d in DirectionLabels[:self._dim]]),
            prefix='V_NB_', suffix=enum[:self._dim])

    def _init_cl_mesh_info(self):
        """Collect mesh info from fields and velocity,
        set and send opencl buffer (self._cl_mesh_info) to device.
        """
        # Space step for fields
        mesh_size = npw.ones(4, dtype=self.gpu_precision)
        mesh_size[:self._dim] = self._reorderVect(
            self.fields_topo.mesh.space_step)

        # Space step for velocity
        self._v_mesh_size = npw.ones(4, dtype=self.gpu_precision)
        self._v_mesh_size[:self._dim] = self._reorderVect(
            self.velocity.topology.mesh.space_step)

        mesh_info = npw.ones((12, ), dtype=self.gpu_precision)
        mesh_info[:4] = mesh_size
        mesh_info[4:8] = self._v_mesh_size
        # Coordinate of the local origin in advection dir.
        mesh_info[8] = self.fields_topo.mesh.origin[self.direction]
        mesh_info[9] = 1. / mesh_size[0]
        mesh_info[10] = 1. / self._v_mesh_size[0]
        assert mesh_size.dtype == self.gpu_precision
        # returns an opencl buffer
        self._cl_mesh_info = cl.Buffer(self.cl_env.ctx, cl.mem_flags.READ_ONLY,
                                       size=mesh_info.nbytes)
        cl.enqueue_write_buffer(self.cl_env.queue,
                                self._cl_mesh_info, mesh_info).wait()

    def _set_work_arrays(self, rwork=None, iwork=None):

        topo = self.fields_on_grid[0].topology
        # Find number and shape of required work arrays
        # For GPU version, no need of numerics works
        # Shape of reference comes from fields, not from velocity
        rwork_length = np.sum([f.nb_components for f in self.fields_on_grid])
        if self.method[Support].find('gpu_2k') >= 0:
            rwork_length += 1  # work array for positions

        # check and/or allocate work arrays according to properties above
        subshape = tuple(topo.mesh.local_resolution)
        self._rwork = WorkSpaceTools.check_work_array(rwork_length, subshape,
                                                      rwork, HYSOP_REAL)

    @abstractmethod
    def global_memory_usage(self, v_shape, shape):
        """Returns an estimation of memory usage

        Parameters
        ----------
        v_shape, shape : tuples
             shapes of the velocity and advected fields
        """
        pass

    def _build_exec_list(self):
        """Prepare GPU kernels sequence
        """
        # Build execution list regarding splitting:
        # Splitting Strang 2nd order:
        #   3D: X(dt/2), Y(dt/2), Z(dt), Y(dt/2), X(dt/2)
        #   2D: X(dt/2), Y(dt), X(dt/2)
        if self.method[Splitting] == 'o2':
            if self._dim == 2:
                self.exec_list = [
                    [self._init_copy, self._compute],  # X(dt/2)
                    [self._init_transpose_xy, self._compute],  # Y(dt)
                    [self._init_transpose_xy, self._compute]  # X(dt/2)
                ]
            elif self._dim == 3:
                self.exec_list = [
                    [self._init_copy, self._compute],  # X(dt/2)
                    [self._init_transpose_xy, self._compute],  # Y(dt/2)
                    [self._init_transpose_xz, self._compute],  # Z(dt)
                    [self._init_transpose_xz, self._compute],  # Y(dt/2)
                    [self._init_transpose_xy, self._compute]  # X(dt/2)
                ]

        # Splitting Strang 2nd order (fullHalf):
        #   X(dt/2), Y(dt/2), Z(dt/2), Z(dt/2), Y(dt/2), X(dt/2)
        elif self.method[Splitting] == 'o2_FullHalf':
            if self._dim == 2:
                self.exec_list = [
                    [self._init_copy, self._compute],  # X(dt/2)
                    [self._init_transpose_xy, self._compute],  # Y(dt)
                    [self._init_copy, self._compute],  # Y(dt)
                    [self._init_transpose_xy, self._compute]  # X(dt/2)
                ]
            elif self._dim == 3:
                self.exec_list = [
                    [self._init_copy, self._compute],  # X(dt/2)
                    [self._init_transpose_xy, self._compute],  # Y(dt/2)
                    [self._init_transpose_xz, self._compute],  # Z(dt/2)
                    [self._init_copy, self._compute],  # Z(dt/2)
                    [self._init_transpose_xz, self._compute],  # Y(dt/2)
                    [self._init_transpose_xy, self._compute]  # X(dt/2)
                ]
        elif self.method[Splitting] == 'x_only':
            self.exec_list = [
                [self._init_copy, self._compute],  # X(dt)
                #[self._init_copy, self._init_copy_r],  # X(dt)
                ]
        else:
            raise ValueError('Splitting type not yet implemeted on GPU: ' +
                             self.method[Splitting])

    def global_memory_usage(self, v_shape, shape):
        if self._use_2_kernels:
            r = (self.velocity.nb_components * v_shape.prod() +
                 (2 * self.fields_on_grid[0].nb_components + 1) * shape.prod())
        else:
            r = (self.velocity.nb_components * v_shape.prod() +
                 2 * self.fields_on_grid[0].nb_components * shape.prod())
        return r * self.cl_env.prec_size

    def _configure_numerical_methods(self):
        pass

    def _buffer_allocations(self):
        """Allocate OpenCL buffers for velocity and advected field.
        """
        # Velocity.
        alloc = not isinstance(self.velocity, OpenClDiscreteField)
        OpenClDiscreteField.fromField(self.cl_env, self.velocity,
                                   self.gpu_precision, simple_layout=False)
        if alloc:
            self.size_global_alloc += self.velocity.mem_size

        # Transported field.
        alloc = not isinstance(self.fields_on_grid[0], OpenClDiscreteField)
        OpenClDiscreteField.fromField(self.cl_env,
                                   self.fields_on_grid[0],
                                   self.gpu_precision,
                                   layout=False)
        if alloc:
            self.size_global_alloc += self.fields_on_grid[0].mem_size

        # Fields on particles
        start = 0
        for f in self.fields_on_grid:
            for i in xrange(start, start + f.nb_components):
                if isinstance(self._rwork[i], np.ndarray):
                    self._rwork[i] = \
                        self.cl_env.global_allocation(self._rwork[i])
            self.fields_on_part[f] = self._rwork[start: start + f.nb_components]
            start += f.nb_components

        if self._use_2_kernels:
            # Particles position
            if isinstance(self._rwork[start], np.ndarray):
                self._rwork[start] = \
                    self.cl_env.global_allocation(self._rwork[start])
            self.part_position = self._rwork[start:start + 1]

        self._work = self.fields_on_part.values()

    def _buffer_initialisations(self):
        """
        OpenCL buffer initializations from user OpenCL kernels.
        Looking for kernels named <code>init<FieldName></code>.
        """
        for gpudf in self.variables:
            match = 'init' + '_'.join(gpudf.name.split('_')[:-1])
            # Looking for initKernel
            if self.prg is not None:
                for k in self.prg.all_kernels():
                    k_name = k.get_info(cl.kernel_info.FUNCTION_NAME)
                    if match.find(k_name) >= 0:
                        if __VERBOSE__:
                            print gpudf.name, '-> OpenCL Kernel', k_name
                        if gpudf == self.velocity:
                            workItemNumber, gwi, lwi = \
                                self.cl_env.get_work_items(self.v_resol_dir)
                        else:
                            workItemNumber, gwi, lwi = \
                                self.cl_env.get_work_items(self.resol_dir)
                        gpudf.setInitializationKernel(OpenClKernelLauncher(
                            cl.Kernel(self.prg, k_name), self.cl_env.queue,
                            gwi, lwi))

    def _collect_kernels_cl_src_copy(self):
        """
        Compile OpenCL sources for copy kernel.
        """
        # # copy settings
        # src, t_dim, b_rows, vec, f_space = self._kernel_cfg['copy']
        # while t_dim > self.resol_dir[0] or (self.resol_dir[0] % t_dim) > 0:
        #     t_dim /= 2
        # gwi, lwi = f_space(self.resol_dir, t_dim, b_rows, vec)

        # # Build code
        # build_options += " -D TILE_DIM_COPY={0}".format(t_dim)
        # build_options += " -D BLOCK_ROWS_COPY={0}".format(b_rows)
        # build_options += self._size_constants
        # prg = self.cl_env.build_src(
        #     src,
        #     build_options,
        #     vec)
        # self.copy = OpenClKernelLauncher(prg.copy,
        #                            self.cl_env.queue, gwi, lwi)
        return OpenClKernelLauncher(cl.enqueue_copy, self.cl_env.queue)

    def _collect_kernels_cl_src_transpositions_xy(self):
        """Compile OpenCL sources for transpositions kernel.

        Notes
        -----

        * Transpositions kernels are launched at initialization.

        This routine sets transpose_xy and transpose_xy_r.
        """
        resol = self.fields_topo.mesh.local_resolution

        # XY transposition settings
        is_XY_needed = self.direction == 1 or self.direction == 0
        if is_XY_needed:
            resol_tmp = resol.copy()
            if self.direction == 1:  # (XY -> YX)
                resol_tmp[0] = resol[1]
                resol_tmp[1] = resol[0]
                ocl_cte = " -D NB_I=NB_Y -D NB_II=NB_X -D NB_III=NB_Z"
            else:
                #  self.direction == 0:  # (YX -> XY) only for sequential
                ocl_cte = " -D NB_I=NB_X -D NB_II=NB_Y -D NB_III=NB_Z"

            self.transpose_xy = self._make_transpose_xy(resol_tmp, ocl_cte)

    def _make_transpose_xy(self, resolution, ocl_cte):
        """Perform xy transposition

        Parameters
        ----------
        resolution : numpy array
            required shape for?
        ocl_cte: string
            compilation options
        """
        build_options = self._size_constants
        src, t_dim, b_rows, is_padding, vec, f_space = \
            self._kernel_cfg['transpose_xy']
        while t_dim > resolution[0] or t_dim > resolution[1] or \
                (resolution[0] % t_dim) > 0 or (resolution[1] % t_dim) > 0:
            t_dim /= 2
        gwi, lwi, blocs_nb = f_space(resolution, t_dim, b_rows, vec)

        if is_padding:
            build_options += " -D PADDING_XY=1"
        else:
            build_options += " -D PADDING_XY=0"
        build_options += " -D TILE_DIM_XY={0}".format(t_dim)
        build_options += " -D BLOCK_ROWS_XY={0}".format(b_rows)
        build_options += " -D NB_GROUPS_I={0}".format(blocs_nb[0])
        build_options += " -D NB_GROUPS_II={0}".format(blocs_nb[1])
        build_options += ocl_cte
        prg = self.cl_env.build_src(src, build_options, vec)
        return OpenClKernelLauncher(prg.transpose_xy, self.cl_env.queue, gwi, lwi)

    def _collect_kernels_cl_src_transpositions_xz(self):
        resol = self.fields_topo.mesh.local_resolution
        resol_tmp = npw.zeros_like(resol)

        is_XZ_needed = self.direction == 2 or self.direction == 1
        # XZ transposition settings
        if is_XZ_needed:
            resol_tmp[...] = resol[...]
            if self.direction == 1:  # ZXY -> YXZ (only for seqential)
                resol_tmp[0] = resol[1]
                resol_tmp[1] = resol[0]
                resol_tmp[2] = resol[2]
                ocl_cte = " -D NB_I=NB_Y -D NB_II=NB_X -D NB_III=NB_Z"
            elif self.direction == 2:
                # YXZ -> ZXY
                resol_tmp[0] = resol[2]
                resol_tmp[1] = resol[0]
                resol_tmp[2] = resol[1]
                ocl_cte = " -D NB_I=NB_Z -D NB_II=NB_X -D NB_III=NB_Y"
                # else:  # XYZ -> ZYX
                #     resol_tmp[0] = resol[2]
                #     resol_tmp[1] = resol[1]
                #     resol_tmp[2] = resol[0]
                #     ocl_cte = " -D NB_I=NB_Z -D NB_II=NB_Y -D NB_III=NB_X"
            self.transpose_xz = self._make_transpose_xz(resol_tmp, ocl_cte)

        # is_XZ_r_needed = self.direction == 2 and self._comm_size > 1
        # if is_XZ_r_needed:
        #     # Reversed XZ transposition settings (ZYX -> XYZ)
        #     resol_tmp[...] = resol[...]
        #     ocl_cte = " -D NB_I=NB_X -D NB_II=NB_Y -D NB_III=NB_Z"
        #     self.transpose_xz_r = self._make_transpose_xz(resol_tmp, ocl_cte)

    def _make_transpose_xz(self, resol_tmp, ocl_cte):

        build_options = self._size_constants
        src, t_dim, b_rows, b_deph, is_padding, vec, f_space = \
            self._kernel_cfg['transpose_xz']

        while t_dim > resol_tmp[0] or t_dim > resol_tmp[2] or \
                (resol_tmp[0] % t_dim) > 0 or (resol_tmp[2] % t_dim) > 0:
            t_dim /= 2
        gwi, lwi, blocs_nb = f_space(resol_tmp, t_dim, b_rows, b_deph, vec)
        if is_padding:
            build_options += " -D PADDING_XZ=1"
        else:
            build_options += " -D PADDING_XZ=0"
        build_options += " -D TILE_DIM_XZ={0}".format(t_dim)
        build_options += " -D BLOCK_ROWS_XZ={0}".format(b_rows)
        build_options += " -D BLOCK_DEPH_XZ={0}".format(b_deph)
        build_options += " -D NB_GROUPS_I={0}".format(blocs_nb[0])
        build_options += " -D NB_GROUPS_III={0}".format(blocs_nb[2])
        build_options += ocl_cte
        prg = self.cl_env.build_src(
            src,
            build_options,
            vec)
        return OpenClKernelLauncher(prg.transpose_xz, self.cl_env.queue, gwi, lwi)

    def _collect_usr_cl_src(self):
        """Build user sources.

        """
        # get extra (opencl kernels) files from method, if any
        user_src = get_extra_args_from_method(self, 'user_src', None)
        if user_src is not None:
            build_options = self._size_constants
            workItemNb, gwi, lwi = self.cl_env.get_work_items(self.resol_dir)
            v_workItemNb, gwi, lwi = self.cl_env.get_work_items(self.v_resol_dir)
            build_options += " -D WI_NB=" + str(workItemNb)
            build_options += " -D V_WI_NB=" + str(v_workItemNb)
            self.prg = self.cl_env.build_src(user_src, build_options, 1)


    def _collect_kernels_cl_src_1k(self):
        """
        Compile OpenCL sources for advection and remeshing kernel.
        """
        build_options = self._size_constants
        src, is_noBC, vec, f_space = self._kernel_cfg['advec_and_remesh']
        gwi, lwi = f_space(self.resol_dir, vec)
        WINb = lwi[0]
        build_options += " -D FORMULA=" + self.method[Remesh].__name__.upper()
        if self._isMultiScale:
            build_options += " -D MS_FORMULA="
            build_options += self.method[MultiScale].__name__.upper()
        if is_noBC:
            build_options += " -D WITH_NOBC=1"
        build_options += " -D WI_NB=" + str(WINb)
        build_options += " -D PART_NB_PER_WI="
        build_options += str(self.resol_dir[0] / WINb)
        # Build code
        src = [s.replace('RKN', self.method[TimeIntegrator].__name__.lower())
               for s in src]
        # Euler integrator
        if self.method[TimeIntegrator] is Euler:
            if not self._isMultiScale:
                src = [s for s in src if s.find(Euler.__name__.lower()) < 0]
                src[-1] = src[-1].replace('advection', 'advection_euler')
        prg = self.cl_env.build_src(
            src, build_options, vec,
            nb_remesh_components=self.fields_on_grid[0].nb_components)

        self.num_advec_and_remesh = OpenClKernelLauncher(
            prg.advection_and_remeshing, self.cl_env.queue, gwi, lwi)

    def _collect_kernels_cl_src_2k(self):
        """
        Compile OpenCL sources for advection and remeshing kernel.
        """
        # Advection
        build_options = self._size_constants
        src, is_noBC, vec, f_space = self._kernel_cfg['advec']
        gwi, lwi = f_space(self.resol_dir, vec)
        WINb = lwi[0]
        if self._isMultiScale:
            build_options += " -D MS_FORMULA="
            build_options += self.method[MultiScale].__name__.upper()
            self._compute_advec = self._compute_advec_multiechelle
        else:
            self._compute_advec = self._compute_advec_simpleechelle

        if is_noBC:
            build_options += " -D WITH_NOBC=1"
        build_options += " -D WI_NB=" + str(WINb)
        build_options += " -D PART_NB_PER_WI="
        build_options += str(self.resol_dir[0] / WINb)
        # Build code
        src = [s.replace('RKN', self.method[TimeIntegrator].__name__.lower())
               for s in src]
        # Adding remeshing weights for the multiscale advection
        if self._isMultiScale:
            src.insert(1, self._kernel_cfg['remesh'][0][1])
        # Euler integrator
        if self.method[TimeIntegrator] is Euler:
            if not self._isMultiScale:
                src = [s for s in src if s.find(Euler.__name__.lower()) < 0]
                src[-1] = src[-1].replace('advection', 'advection_euler')
                self._compute_advec = self._compute_advec_euler_simpleechelle
        prg = self.cl_env.build_src(
            src,
            build_options,
            vec,
            nb_remesh_components=self.fields_on_grid[0].nb_components)

        self.num_advec = OpenClKernelLauncher(
            prg.advection_kernel, self.cl_env.queue, gwi, lwi)

        # remeshing
        build_options = self._size_constants
        src, is_noBC, vec, f_space = self._kernel_cfg['remesh']
        gwi, lwi = f_space(self.resol_dir, vec)
        WINb = lwi[0]

        build_options += " -D FORMULA=" + self.method[Remesh].__name__.upper()
        if is_noBC:
            build_options += " -D WITH_NOBC=1"
        build_options += " -D WI_NB=" + str(WINb)
        build_options += " -D PART_NB_PER_WI="
        build_options += str(self.resol_dir[0] / WINb)
        # Build code
        prg = self.cl_env.build_src(
            src, build_options, vec,
            nb_remesh_components=self.fields_on_grid[0].nb_components)
        self.num_remesh = OpenClKernelLauncher(
            prg.remeshing_kernel, self.cl_env.queue, gwi, lwi)

    @debug
    @profile
    def apply(self, simulation, dt_coeff, split_id, old_dir):
        """
        Apply operator along specified splitting direction.
        @param t : Current time
        @param dt : Time step
        @param d : Splitting direction
        @param split_id : Splitting step id
        """
        # If first direction of advection, wait for work gpu fields
        # It avoid wait_for lists to increase indefinitely
        # In practice, all events are terminated so wait() resets events list
        if split_id == 0:
            for v in self.fields_on_grid + [self.velocity]:
                v.clean_events()
        for exe in self.exec_list[split_id]:
            exe(simulation, dt_coeff, split_id, old_dir)

    def _init_copy(self, simulation, dt_coeff, split_id, old_dir):
        wait_evt = self.fields_on_grid[0].events
        for g, p in zip(self.fields_on_grid[0].gpu_data,
                        self.fields_on_part[self.fields_on_grid[0]]):
            evt = self.copy.launch_sizes_in_args(p, g, wait_for=wait_evt)
            #evt = self.copy(g, p, wait_for=wait_evt)
            self._init_events[self.fields_on_grid[0]].append(evt)

    # def _init_copy_r(self, simulation, dt_coeff, split_id, old_dir):
    #     wait_evt = self.fields_on_grid[0].events
    #     for g, p in zip(self.fields_on_grid[0].gpu_data,
    #                  self.fields_on_part[self.fields_on_grid[0]]):
    #         evt = self.copy.launch_sizes_in_args(g, p, wait_for=wait_evt)
    #         #evt = self.copy(p, g, wait_for=wait_evt)
    #         self._init_events[self.fields_on_grid[0]].append(evt)

    def _init_transpose_xy(self, simulation, dt_coeff, split_id, old_dir):
        wait_evt = self.fields_on_grid[0].events
        for g, p in zip(self.fields_on_grid[0].gpu_data,
                        self.fields_on_part[self.fields_on_grid[0]]):
            evt = self.transpose_xy(g, p, wait_for=wait_evt)
            self._init_events[self.fields_on_grid[0]].append(evt)

    # def _init_transpose_xy_r(self, simulation, dt_coeff, split_id, old_dir):
    #     wait_evt = self.fields_on_grid[0].events
    #     for g, p in zip(self.fields_on_grid[0].gpu_data,
    #                     self.fields_on_part[self.fields_on_grid[0]]):
    #         evt = self.transpose_xy_r(p, g, wait_for=wait_evt)
    #         self._init_events[self.fields_on_grid[0]].append(evt)

    def _init_transpose_xz(self, simulation, dt_coeff, split_id, old_dir):
        wait_evt = self.fields_on_grid[0].events
        for g, p in zip(self.fields_on_grid[0].gpu_data,
                        self.fields_on_part[self.fields_on_grid[0]]):
            evt = self.transpose_xz(g, p, wait_for=wait_evt)
            self._init_events[self.fields_on_grid[0]].append(evt)

    # def _init_transpose_xz_r(self, simulation, dt_coeff, split_id, old_dir):
    #     wait_evt = self.fields_on_grid[0].events
    #     for g, p in zip(self.fields_on_grid[0].gpu_data,
    #                     self.fields_on_part[self.fields_on_grid[0]]):
    #         evt = self.transpose_xz_r(p, g, wait_for=wait_evt)
    #         self._init_events[self.fields_on_grid[0]].append(evt)

    def _compute_advec_euler_simpleechelle(self, simulation, dt_coeff, split_id, old_dir):
        dt = simulation.time_step * dt_coeff
        wait_evts = self.velocity.events + \
            self._init_events[self.fields_on_grid[0]]
        # Advection
        evt = self.num_advec(
            self.velocity.gpu_data[self.direction],
            self.part_position[0],
            self.gpu_precision(dt),
            self._cl_mesh_info,
            wait_for=wait_evts)
        self._init_events[self.fields_on_grid[0]].append(evt)

    def _compute_advec_simpleechelle(self, simulation, dt_coeff, split_id, old_dir):
        dt = simulation.time_step * dt_coeff
        wait_evts = self.velocity.events + \
            self._init_events[self.fields_on_grid[0]]
        # Advection
        evt = self.num_advec(
            self.velocity.gpu_data[self.direction],
            self.part_position[0],
            self.gpu_precision(dt),
            self._cl_mesh_info,
            wait_for=wait_evts)
        self._init_events[self.fields_on_grid[0]].append(evt)

    def _compute_advec_multiechelle(self, simulation, dt_coeff, split_id, old_dir):
        dt = simulation.time_step * dt_coeff
        wait_evts = self.velocity.events + \
            self._init_events[self.fields_on_grid[0]]
        # Advection
        evt = self.num_advec(
            self.velocity.gpu_data[self.direction],
            self.part_position[0],
            self.gpu_precision(dt),
            self.gpu_precision(1. / self._v_mesh_size[1]),
            self.gpu_precision(1. / self._v_mesh_size[2]),
            self._cl_mesh_info,
            wait_for=wait_evts)
        self._init_events[self.fields_on_grid[0]].append(evt)

    def _compute_2k(self, simulation, dt_coeff, split_id, old_dir):
        self._compute_advec(simulation, dt_coeff, split_id, old_dir)
        wait_evts = self._init_events[self.fields_on_grid[0]] + \
            self.fields_on_grid[0].events
        nbc = self.fields_on_grid[0].nb_components
        evt = self.num_remesh(*tuple(
            [self.part_position[0], ] +
            [self.fields_on_part[self.fields_on_grid[0]][i]
             for i in xrange(nbc)] +
            [self.fields_on_grid[0].gpu_data[i] for i in xrange(nbc)] +
            [self._cl_mesh_info, ]),
                              wait_for=wait_evts)
        self.fields_on_grid[0].events.append(evt)
        self._init_events[self.fields_on_grid[0]] = []

    def _compute_1k_multiechelle(self, simulation, dt_coeff, split_id, old_dir):
        if split_id==0 and self._synchronize is not None:
            self._synchronize(self.velocity.data)
            self.velocity.toDevice()
        dt = simulation.time_step * dt_coeff
        wait_evts = self.velocity.events + \
            self._init_events[self.fields_on_grid[0]] + \
            self.fields_on_grid[0].events
        nbc = self.fields_on_grid[0].nb_components
        evt = self.num_advec_and_remesh(*tuple(
            [self.velocity.gpu_data[self.direction], ] +
            [self.fields_on_part[self.fields_on_grid[0]][i]
             for i in xrange(nbc)] +
            [self.fields_on_grid[0].gpu_data[i] for i in xrange(nbc)] +
            [self.gpu_precision(dt),
             self.gpu_precision(1. / self._v_mesh_size[1]),
             self.gpu_precision(1. / self._v_mesh_size[2]),
             self._cl_mesh_info]),
                                        wait_for=wait_evts)
        self.fields_on_grid[0].events.append(evt)
        self._init_events[self.fields_on_grid[0]] = []

    def _compute_1k_simpleechelle(self, simulation, dt_coeff, split_id, old_dir):
        dt = simulation.time_step * dt_coeff
        wait_evts = self.velocity.events + \
            self._init_events[self.fields_on_grid[0]] + \
            self.fields_on_grid[0].events
        nbc = self.fields_on_grid[0].nb_components
        evt = self.num_advec_and_remesh(*tuple(
            [self.velocity.gpu_data[self.direction], ] +
            [self.fields_on_part[self.fields_on_grid[0]][i]
             for i in xrange(nbc)] +
            [self.fields_on_grid[0].gpu_data[i] for i in xrange(nbc)] +
            [self.gpu_precision(dt), self._cl_mesh_info]),
                                        wait_for=wait_evts)
        self.fields_on_grid[0].events.append(evt)
        self._init_events[self.fields_on_grid[0]] = []


    def _compute_1k_euler_simpleechelle(self, simulation, dt_coeff, split_id, old_dir):
        dt = simulation.time_step * dt_coeff
        wait_evts = self.velocity.events + \
            self._init_events[self.fields_on_grid[0]] + \
            self.fields_on_grid[0].events
        nbc = self.fields_on_grid[0].nb_components
        evt = self.num_advec_and_remesh(*tuple(
            [self.velocity.gpu_data[self.direction], ] +
            [self.fields_on_part[self.fields_on_grid[0]][i]
             for i in xrange(nbc)] +
            [self.fields_on_grid[0].gpu_data[i] for i in xrange(nbc)] +
            [self.gpu_precision(dt), self._cl_mesh_info]),
                                        wait_for=wait_evts)
        self.fields_on_grid[0].events.append(evt)
        self._init_events[self.fields_on_grid[0]] = []

    def get_profiling_info(self):
        for k in [self.copy, self.transpose_xy, self.transpose_xy_r,
                  self.transpose_xz, self.transpose_xz_r,
                  self.num_advec_and_remesh,
                  self.num_advec, self.num_remesh]:
            if k is not None:
                for p in k.profile:
                    self.profiler += p

    @debug
    def finalize(self):
        """
        Cleaning, if required.
        """
        pass
        # for w in self._rwork:
        #     self.cl_env.global_deallocation(w)
        # self.cl_env.global_deallocation(self._cl_mesh_info)
