"""Interpolation methods

.. currentmodule hysop.numerics

* :class:`~interpolation.Linear`

"""
from hysop.constants import HYSOP_INTEGER, EPS
from hysop.tools.misc import WorkSpaceTools
import numpy as np
from hysop.tools.numpywrappers import npw
from abc import ABCMeta
try:
    from hysop.f2hysop import interpolation as finterpol
except:
    finterpol = None


class Interpolation(object):
    """Common interface to all 'interpolation-like' numerical methods
    (interpolation, remesh)
    """
    __metaclass__ = ABCMeta
    ghosts_layer_size = 0

    def __init__(self, topo_source, direction,
                 topo_target=None, rwork=None, iwork=None):
        """"Common interface to all 'interpolation-like' numerical methods
        (interpolation, remesh)

        Parameters
        -----------
        topo_source, topo_target : :class:`~hysop.topology.topology.CartesianTopology`
            mpi process and data distribution for source and target
            (see Notes for details)
        direction : int
            direction of interpolation
        rwork, iwork : list of numpy arrays
            internal work spaces for real (work) and integer (iwork) data.

        Attributes
        ----------
        ghosts_layer_size : int, static
            minimal number of points required inside the ghost layer
            for this scheme.

        Notes
        -----
        * 'source' concerned all data related to the grid on which the field
        to interpolate is known, and 'target' is used for data related
        to the grid on which the field values must be computed using
        interpolation.
        * topo_target is required if and only if source and target grid
        have different resolution.
        """
        self.direction = direction
        self.topo_source = topo_source
        if topo_target is None:
            topo_target = topo_source
        self.topo_target = topo_target
        # Check ghost layer size (min required in the interpolation direction)
        msg = 'Ghost layer is too small for the chosen scheme.'
        required_ghost_layer = self.ghosts_layer_size
        assert topo_source.ghosts()[direction] >= required_ghost_layer, msg

        # work arrays
        self._rwork, self._iwork = self._set_work_arrays(
            rwork=rwork, iwork=iwork)

        # indices of grid points that will be remeshed:
        # - at most the resolution in remeshed direction
        #  --> update during call,
        # (for example, it may depend on the number of particles)
        # - all grid points except ghosts in other directions
        ic = self.topo_source.mesh.compute_index
        # grid indices in 'source' grid
        self._icoords = np.ix_(
            *(np.arange(ic[i].start, ic[i].stop)
              for i in xrange(topo_source.domain.dim)))
        self._inv_dx = 1. / self.topo_source.mesh.space_step

        #source_res = topo_source.mesh.local_resolution
        #target_res = topo_target.mesh.local_resolution
        #msg = 'Interpolation error : only implemented in 3D.'
        #assert topo_source.domain.dim == 3, msg
        #assert topo_target.domain.dim == 3, msg
        #if (source_res == target_res).all():
        #    msg = 'Interpolation error : the two grids have the same '
        #    msg += 'resolution, interpolation is useless.'
        #shape_cfc = (source_res[0], target_res[2], source_res[1])
        #shape_ccf = (source_res[0], source_res[1], target_res[2])
        #shape_ccc = (source_res[0], source_res[2], source_res[1])
        #shape_ccf2 = (source_res[0], source_res[2], target_res[2])
        #self._rwork = [npw.zeros(shape_cfc), npw.zeros(shape_ccf),
        #               npw.zeros(shape_ccc), npw.zeros(shape_ccf2)]
        self.hsource = topo_source.mesh.space_step
        self.htarget = topo_target.mesh.space_step
        self.topo_target.build_subcomms()
        self.sub_comms = self.topo_target.sub_comms
        self.neighbours = self.topo_target.sub_neighbours
        if finterpol is not None:
            finterpol.initialize_interpolation_kernel('L4_4')

        # indices of points on which interpolation occurs.
        # This corresponds to the mesh of topo_target
        # but is updated at each call, depending on the
        # size of 'positions' argument and of the direction of interpolation.
        self._target_indices = list(self.topo_target.mesh.compute_index)

    def reset_target_indices(self, new_indices):
        """Reset list of indices of points on which interpolation occurs.
        Only useful when 'positions' argument shape is not equal
        to topo_target.mesh shape, e.g. when a threshold is used
        to initialize particles.

        Parameters
        ----------
        new_indices : a list of slices
            a slice per direction to define the range of
            indices of interest. Like topo.mesh.compute_index attribute.
        """
        self._target_indices = list(new_indices)

    def _set_work_arrays(self, rwork=None, iwork=None):
        """Check and/or allocate internal work buffers.

        Parameters
        ----------
        rwork, iwork : list of numpy arrays
            internal work spaces for real (rwork) and integer (iwork) data.

        """
        wk_prop = self.get_work_properties(self.topo_target)
        noghosts_shape = (np.prod(self.topo_target.mesh.compute_resolution),)
        if wk_prop['rwork'] is None:
            result_r = []
        else:
            result_r = WorkSpaceTools.check_work_array(len(wk_prop['rwork']),
                                                       noghosts_shape, rwork)
        if wk_prop['iwork'] is None:
            result_i = []
        else:
            result_i = WorkSpaceTools.check_work_array(len(wk_prop['iwork']),
                                                       noghosts_shape, iwork,
                                                       data_type=HYSOP_INTEGER)
        return result_r, result_i

    @staticmethod
    def get_work_properties(topo):
        """Get properties of internal work arrays.

        Returns
        -------
        dictionnary
           keys = 'rwork' and 'iwork', values = list of shapes of internal
           arrays required by this method (real arrays for rwork, integer
           arrays for iwork).

        Parameters
        ----------
        topo : :class:`hysop.topology.topology.CartesianTopology`
            topology on which the method will be applied.
        """
        # default is None, None.
        # This function must be reimplemented in derived classes
        # (but we cannot combine abstract and static in python2.7 ...)
        return {'rwork': None, 'iwork': None}

    def _assert_target_in_source(self, target_coords):
        """Check if target grid is strictly included
        into source grid

        This is required to be sure that 'left' and 'right'
        points exist for each point on target grid.
        This obviously depends on the interpolation order.
        """
        source_coords = self.topo_source.mesh.coords[self.direction]
        assert np.min(target_coords) >= source_coords.flat[0]
        assert np.max(target_coords) < source_coords.flat[-1]
        return True

    def _compute_iy(self, positions, work):
        """Compute distance between points to be interpolated and their
        closest point on the left.

        Parameters
        ----------
        positions : numpy array
            points to be interpolated. positions shape == topo.mesh.local_resolution
            except in direction of interpolation where any number of points
            is allowed, with a max of points equal to the number of grid
            points in this direction (excluding ghosts).
        work : numpy array.
            Used as local buffer. work must be of the same shape
            as positions and contains iy in return,
            i_y = (xp - xg)/dx

        Returns
        -------
        left_points_indices : list of numpy arrays or slices
            indices in source grid of neighbours (left) points for
            each point in positions.

        Usage : field[left_points_indices] will return the array of
        'left' points (in current direction)
        while field[target_indices] = current points

        Remark
        ------
        No memory allocation in this function --> everything is taken
        from rwork or iwork
        """
        # local origin and space step in current direction
        x0 = self.topo_source.mesh.origin[self.direction]
        #i_y = work
        # memory optim: check if input work 'belongs' to rwork
        #assert npw.arrays_share_data(i_y, work)
        # -- set indices of closest grid points, on the left --
        # Initialized with source grid coordinates
        left_points_indices = list(self._icoords)
        # = [self._icoords[i] for i in xrange(dimension)]
        # pick memory from self._iwork
        iwork_dir = WorkSpaceTools.check_work_array(
            1, work.shape, self._iwork, data_type=HYSOP_INTEGER)
        # we need an array (filled later) for current direction
        left_points_indices[self.direction] = iwork_dir[0]
        # check memory consumption
        assert npw.arrays_share_data(left_points_indices[self.direction],
                                     self._iwork[0])
        # -- compute distance between particles positions and
        # closest grid point, on the left --
        work[...] = (positions[self._target_indices] - x0) *\
            self._inv_dx[self.direction]
        #self._assert_particles_roundoff(work, self._target_indices)
        # update interpolation point indices (in self._iwork)
        left_points_indices[self.direction][...] = \
            npw.asintegerarray(np.floor(work[...]))
        work[...] -= left_points_indices[self.direction][...]
        nmax = self.topo_source.mesh.local_resolution[self.direction]
        # Ensure all points belongs to local domain
        # (including ghosts)
        assert left_points_indices[self.direction].max() < nmax
        assert left_points_indices[self.direction].min() >= 0
        return left_points_indices

    @staticmethod
    def _assert_particles_roundoff(work, ic):
        """trick to avoid case where particle is very close to
        grid point and 'floor' result in wrong grid
        point because of numerical eps.
        For example, if xp - xi == -1e-15, floor(xp)=xi-1
        which is not what is expected.

        Done in 'assertion' --> only in debug mode.
        """
        work[1][ic] = np.ceil(work[0][ic])
        ilist = np.where(work[1][ic] - work[0][ic] < 10. * EPS)
        work[0][ic][ilist] = work[1][ic][ilist]
        return True

    def _xy(self, vin, vout):
        """Interpolate a field along X and Y-axis

        Parameters
        ----------
        vin : list of numpy arrays
            field on the source grid
        vout : list of numpy arrays
            field on the target/source grid
        Notes
        -----
        * vout and vin must have the same resolution along
          the first direction.
        """
        # Permutation to prepare first interpolation
        for i in xrange(3):
            self._rwork[2][:, i, :] = vin[:, :, i]
        # Interpolation along last dir = Y
        # in : vpermut(nc_x, nc_z, nc_y)
        # out : vm(nc_x, nf_y, nc_z)
        finterpol.interpol_lastdir_permut(self._rwork[2], self.hsource[:2],
                                          self._rwork[3], self.htarget[:2],
                                          self.neighbours[:, YDIR],
                                          self.sub_comms[YDIR])
        # Interpolation along X
        # in :  vm(nc_x, nf_y, nc_z)
        # out : vout(nf_x, nf_y, nc_z)
        finterpol.interpol_firstdir_no_com(self._rwork[3],
                                           self.hsource[:2],
                                           vout, self.htarget[:2])

    def _yz(self, vin, vout):
        """Interpolate a field along Y and Z-axis

        Parameters
        ----------
        vin : list of numpy arrays
            field on the source grid
        vout : list of numpy arrays
            field on the target/source grid
        Notes
        -----
        * vout and vin must have the same resolution along
          the first direction.
        """
        # Interpolation along Z + permutation between Y and Z
        # in : vin(nc_x, nc_y, nc_z)
        # out : vtmp(nc_x, nf_z, nc_y)
        finterpol.interpol_lastdir_permut(vin, self.hsource[1:],
                                          self._rwork[0], self.htarget[1:],
                                          self.neighbours[:, ZDIR],
                                          self.sub_comms[ZDIR])
        # Interpolation along Y
        # (=third direction thanks to previous permutation)
        # + permutation between Y and Z
        # in : vtmp(nc_x, nf_z, nc_y)
        # out : vout(nc_x, nf_y, nf_z)
        finterpol.interpol_lastdir_permut(self._rwork[0], self.hsource[XDIR],
                                          vout, self.htarget[XDIR],
                                          self.neighbours[YDIR, :],
                                          self.sub_comms[YDIR])

    def _xz_permut(self, vin, vout):
        """Interpolate a field along X and Z-axis, result order = Y,X,Z

        Parameters
        ----------
        vin : list of numpy arrays
            field on the source grid
        vout : list of numpy arrays
            field on the target grid

        Notes
        -----
        * vout and vin must have the same resolution along
          the first direction.
        """
        # Interpolation along Z
        # Interpolation along Z
        # in : vin(nc_x, nc_y, nc_z)
        # out : vtmp(nc_x, nc_y, nf_z)
        finterpol.interpol_lastdir(vin, self.hsource[::2],
                                   self._rwork[1], self.htarget[::2],
                                   self.neighbours[:, ZDIR],
                                   self.sub_comms[ZDIR])
        # Interpolation along X
        # in : vtmp(nc_x, nc_y, nf_z)
        # out : vout(nc_y, nf_x, nf_z)
        finterpol.interpol_firstdir_permut_no_com(self._rwork[1],
                                                  self.hsource[::2],
                                                  vout, self.htarget[::2])

    def _xy_permut(self, vin, vout):
        """Interpolate a field along X and Y-axis, result order = Z,X,Y

        Parameters
        ----------
        vin : list of numpy arrays
            field on the source grid
        vout : list of numpy arrays
            field on the target grid

        Notes
        -----
        * vout and vin must have the same resolution along
          the first direction.
        """
        # Permutation to prepare first interpolation
        for i in xrange(3):
            self._rwork[2][:, i, :] = vin[:, :, i]
        # Interpolation along last dir = Y
        # in : vpermut(nc_x, nc_z, nc_y)
        # out : vm(nc_x, nc_z, nf_y)
        finterpol.interpol_lastdir(self._rwork[2], self.hsource[:2],
                                   self._rwork[3], self.htarget[:2],
                                   self.neighbours[:, YDIR],
                                   self.sub_comms[YDIR])
        # Interpolation along X
        # in :  vm(nc_x, nc_z, nf_y)
        # out : vout(nc_z, nf_x, nf_y)
        finterpol.interpol_firstdir_permut_no_com(self._rwork[3],
                                                  self.hsource[:2],
                                                  vout, self.htarget[:2])

    def _2d_3d_vect(self, vin, vout):
        """Interpolate each component of a vector along a transverse direction
        """
        # For Vx, interpolation along Y and Z
        self._yz(vin[XDIR], vtmp[XDIR])
        # For Vy, interpolation along Z (with communications)
        # then along X (no communication required)
        self._xz_permut(vin[YDIR], vtmp[YDIR])
        # For Vz, interpolation along Y (with communications)
        # then along X (no communication required)
        self._xy_permut(vin[ZDIR], vtmp[ZDIR])

        finterpol.interpol_firstdir_no_com(vtmp[XDIR], self.hsource[XDIR],
                                           vout[XDIR], self.htarget[XDIR])
        finterpol.interpol_firstdir(vtmp[YDIR], self.hsource[YDIR],
                                    vout[YDIR], self.htarget[YDIR],
                                    self.neighbours[:, YDIR],
                                    self.sub_comms[YDIR])
        finterpol.interpol_firstdir(vtmp[ZDIR], self.hsource[ZDIR],
                                    vout[ZDIR], self.htarget[ZDIR],
                                    self.neighbours[:, ZDIR],
                                    self.sub_comms[ZDIR])

    def _3d(self, vin, vout):
        """3D interpolation of a field to a targetr grid - no transpositions.
        """
        # rwork = vtmp(nf_x, nc_y, nc_z)
        finterpol.interpol_firstdir_no_com(vin, self.hsource[XDIR],
                                           self._rwork[0], self.htarget[XDIR])
        self._yz(self._rwork[0], vout)


class Linear(Interpolation):
    """Linear interpolation of a field"""

    ghosts_layer_size = 1

    def __init__(self, interpolated_field, **kwds):
        """
        Parameters
        ----------
        interpolated_field : numpy array
            data to be interpolated
        kwds : base class parameters

        Notes
        -----
        * topo argument corresponds to the topology of field interpolated_field
        * it would be probably better to pass interpolated_field
          during __call__ but interpolation is used as rhs in odesolvers and
          call function must fit with rhs arg list.

        """
        self.interpolated_field = interpolated_field
        super(Linear, self).__init__(**kwds)

    @staticmethod
    def get_work_properties(topo):
        """Get properties of internal work arrays.

        Returns
        -------
        dictionnary
           keys = 'rwork' and 'iwork', values = list of shapes of internal
           arrays required by this method (real arrays for rwork, integer
           arrays for iwork).

        Parameters
        ----------
        topo : :class:`hysop.topology.topology.CartesianTopology`
            topology on which the method will be applied.
        """
        noghosts_shape = (np.prod(topo.mesh.compute_resolution),)
        return {'rwork': [noghosts_shape, ], 'iwork': [noghosts_shape, ]}

    def __call__(self, t, positions, result):
        """interpolate values at nodes positions using linear interpolation.

        Parameters
        ----------
        t : double
            current time.
        positions : list of numpy arrays
            coordinates of points where to interpolate
        result : list of numpy arrays

        Returns
        -------
        result

        Notes
        -----
        * since interpolation function is likely to be used
        as right-hand side of an ode solver, the list of arguments must fit
        with the integrator requirements, which explains the 't' argument
        unused below but required, or the positions and result as lists.
        """
        assert isinstance(positions, list)
        assert isinstance(result, list)
        assert result[0].shape == positions[0].shape
        assert (self.interpolated_field.shape ==
                self.topo_source.mesh.local_resolution).all()
        #assert self._assert_target_in_source(positions)
        # linear interpolation:
        # input = interpolated_field (field values at points xi),
        #         pos, a vector of p points
        # output = res(pos), interpolation of interpolated_field at points pos
        # intf =interpolated_field
        # res(pos_p) = (pos_p - xi)/h . intf_i+1 + (x_i+1 - pos_p)/h . intf_i
        #          = iy.intf_i+1 + (1 - iy).intf_i
        # where iy = (pos_p - x0)/h - i, i = floor((pos_p - x0)/h)

        # number of points to interpolate
        nb_points = positions[0].shape[self.direction]
        # indices of points where we need interpolation
        self._target_indices[self.direction] = slice(0, nb_points)
        # pick buffer memory from self._rwork
        # wk[0] shares memory with rwork[0] but data are rearranged
        # according to positions[0].shape.
        print nb_points, self.direction
        wk = WorkSpaceTools.check_work_array(
            1, positions[0][self._target_indices].shape, self._rwork)
        # memory optim: check if input work 'belongs' to rwork
        assert npw.arrays_share_data(self._rwork[0], wk[0])
        left_points_indices = self._compute_iy(positions[0], wk[0])
        assert npw.arrays_share_data(left_points_indices[self.direction],
                                     self._iwork[0])
        print self.interpolated_field[left_points_indices].shape
        #refsize = positions[0].size
        #refshape = positions[0].shape
        #wk.append(result[0][])
        #res = WorkSpaceTools.check_work_array(1, refshape, result)
        result[0][self._target_indices] = \
            self.interpolated_field[left_points_indices] * (1. - wk[0][...])
        # compute 'right points' indices
        left_points_indices[self.direction] += 1

        result[0][self._target_indices] += \
            self.interpolated_field[left_points_indices] * wk[0][...]
        return result
