"""Test remesh schemes
"""
from hysop.numerics.remeshing import Linear, L2_1, Remeshing
from hysop.tools.numpywrappers import npw
from hysop.constants import EPS
from hysop import Box, Discretization
import numpy as np
cos = np.cos

disc = [None, ] * 3
g = 2
nx = 65
ny = 49
nz = 57
disc[0] = Discretization([nx, ], [g, ])
disc[1] = Discretization([nx, ny], [g, g])
disc[2] = Discretization([nx, ny, nz], [g, g, g])
doms = [None, ] * 3
topos = [None, ] * 3
rworks = [None, ] * 3
iworks = [None, ] * 3
for case in xrange(3):
    dim = case + 1
    doms[case] = Box(origin=[0.] * dim, length=[4. * np.pi, ] * dim)
    topos[case] = doms[case].create_topology(disc[case])
    refsize = (np.prod(topos[case].mesh.local_resolution) + 324, )
    iworks[case] = [npw.int_zeros(refsize)]
    rworks[case] = [npw.zeros(refsize), npw.zeros(refsize)]
dx = [0., 0. , 0.] # topos[2].mesh.space_step


def run_remesh(kernel, dimension, velo=0., with_work=False):
    """Interpolate a field and compare with reference
    (cosine function), for a given domain dimension
    """
    topo = topos[dimension - 1]
    rwork, iwork = None, None
    if with_work:
        rwork = rworks[dimension - 1]
        iwork = iworks[dimension - 1]
    ind = topo.mesh.compute_index
    lenfield = 2
    fieldp = [None, ] * lenfield
    result = [None, ] * lenfield
    reference = [None, ] * lenfield
    for direction in xrange(topo.domain.dim):
        h = topo.mesh.space_step[direction]
        x = topo.mesh.coords[direction]
        reference = [npw.zeros(topo.mesh.local_resolution)
                     for _ in xrange(lenfield)]
        reference[0][...] = np.cos(x)
        if lenfield > 1:
            reference[1][...] = np.sin(x)
        shape = topo.mesh.local_resolution.copy()
        shape[direction] -= 2 * g
        # note : this should work for any number of points
        # in direction, as soon as work array is large enough
        # + EPS to be sure that particles are (almost) on grid
        # but on the right of the grid point when velo=0
        positions = npw.random(shape) * velo
        positions[...] += 10 * EPS
        positions[...] += topo.mesh.compute_coords[direction]
        for i in xrange(lenfield):
            fieldp[i] = npw.zeros_like(positions)
            result[i] = npw.zeros_like(reference[i])
        fieldp[0][...] = np.cos(positions)
        if lenfield > 1:
            fieldp[1][...] = np.sin(positions)
        remesh = Remeshing(kernel, topo_source=topo, direction=direction,
                           rwork=rwork, iwork=iwork)
        result = remesh(positions, fieldp, result)
        tol = 0.5 * h ** 2
        #plt.plot(x.flat[2:-2], result[ind], 'o-', x.flat[2:-2], reference[ind], '+-')
        #plt.show()
        for i in xrange(lenfield):
            assert np.allclose(result[i][ind], reference[i][ind], rtol=tol)


def test_linear_remesh_1d_velocity_null():
    """Linear remesh on 1d domain"""
    run_remesh(Linear, 1)


def test_linear_remesh_1d():
    """Linear remesh on 1d domain"""
    run_remesh(Linear, 1, 0.3 * dx[0])


def test_linear_remesh_2d_velocity_null():
    """Linear remesh on 2d domain"""
    run_remesh(Linear, 2)


def test_linear_remesh_2d():
    """Linear remesh on 2d domain"""
    run_remesh(Linear, 2, 0.3 * dx[1])


def test_linear_remesh_3d_velocity_null():
    """Linear remesh on 3d domain"""
    run_remesh(Linear, 3)


def test_linear_remesh_3d():
    """Linear remesh on 3d domain"""
    run_remesh(Linear, 3, 0.3 * dx[2])


def test_linear_remesh_1d_wk_velocity_null():
    """Linear remesh on 1d domain, external work arrays"""
    run_remesh(Linear, 1, 0., True)


def test_linear_remesh_1d_wk():
    """Linear remesh on 1d domain, external work arrays"""
    run_remesh(Linear, 1, 0.3 * dx[0], True)


def test_linear_remesh_2d_wk_velocity_null():
    """Linear remesh on 2d domain, external work arrays"""
    run_remesh(Linear, 2, 0., True)


def test_linear_remesh_2d_wk():
    """Linear remesh on 2d domain, external work arrays"""
    run_remesh(Linear, 2, 0.3 * dx[1], True)


def test_linear_remesh_3d_wk_velocity_null():
    """Linear remesh on 3d domain, external work arrays"""
    run_remesh(Linear, 3, 0., True)


def test_linear_remesh_3d_wk():
    """Linear remesh on 3d domain, external work arrays"""
    run_remesh(Linear, 3, 0.3 * dx[2], True)


def test_l21_remesh_1d_velocity_null():
    """L2_1 remesh on 1d domain"""
    run_remesh(L2_1, 1)


def test_l21_remesh_1d():
    """L2_1 remesh on 1d domain"""
    run_remesh(L2_1, 1, 0.3 * dx[0])


def test_l21_remesh_2d_velocity_null():
    """L2_1 remesh on 2d domain"""
    run_remesh(L2_1, 2)


def test_l21_remesh_2d():
    """L2_1 remesh on 2d domain"""
    run_remesh(L2_1, 2, 0.3 * dx[1])


def test_l21_remesh_3d_velocity_null():
    """L2_1 remesh on 3d domain"""
    run_remesh(L2_1, 3)


def test_l21_remesh_3d():
    """L2_1 remesh on 3d domain"""
    run_remesh(L2_1, 3, 0.3 * dx[2])


def test_l21_remesh_1d_wk_velocity_null():
    """L2_1 remesh on 1d domain, external work arrays"""
    run_remesh(L2_1, 1, 0., True)


def test_l21_remesh_1d_wk():
    """L2_1 remesh on 1d domain, external work arrays"""
    run_remesh(L2_1, 1, 0.3 * dx[0], True)


def test_l21_remesh_2d_wk_velocity_null():
    """L2_1 remesh on 2d domain, external work arrays"""
    run_remesh(L2_1, 2, 0., True)


def test_l21_remesh_2d_wk():
    """L2_1 remesh on 2d domain, external work arrays"""
    run_remesh(L2_1, 2, 0.3 * dx[1], True)


def test_l21_remesh_3d_wk_velocity_null():
    """L2_1 remesh on 3d domain, external work arrays"""
    run_remesh(L2_1, 3, 0., True)


def test_l21_remesh_3d_wk():
    """L2_1 remesh on 3d domain, external work arrays"""
    run_remesh(L2_1, 3, 0.3 * dx[2], True)
