"""Update ghost points of a list of numpy arrays
for a given topology.

Setup for send/recv process of ghosts points for a list
of numpy arrays, for a given topology.

.. currentmodule:: hysop.numerics.update_ghosts

* :class:`~UpdateGhosts` : udpate ghosts layers in a 'classical' way
* :class:`~UpdateGhostsFull` : update also points in the corners of the grid,
i.e. at the intersection of ghosts points lines.

Usage :

.. code::

    # init
    up = UpdateGhosts(topo, 2)
    ...
    # update
    up([field1, field2])

"""
from hysop.constants import debug, PERIODIC, hysop.core.mpi_REAL
from hysop.tools.numpywrappers import npw
import numpy as np


class UpdateGhosts(object):
    """Ghost points synchronization for a list of numpy arrays
    """

    @debug
    def __init__(self, topo, nb_elements):
        """
        Parameters
        ----------
        topo : :class:`hysop.topology.topology.CartesianTopology`
            data and mpi process distribution
        nb_elements : int
            number of arrays that will be update
            at each call.

        Notes
        ------
        * nb_elements and topo.mesh.local_resolution will be used to
          allocate memory for local buffers used for send-recv process.
        """
        # The mpi topology and mesh distribution
        self.topology = topo
        # Ghost layer
        self.ghosts = self.topology.mesh.discretization.ghosts
        # Indices of points to be filled from previous neighbour
        # Each component is a slice and corresponds to a direction in
        # the topology grid.
        self._g_fromprevious = []
        # Indices of points to be filled from next neighbour.
        self._g_fromnext = []
        # Indices of points to be sent to next neighbour.
        self._g_tonext = []
        # Indices of points to be sent to previous neighbour.
        self._g_toprevious = []
        # Buffers that contains ghosts points to be sent
        self._sendbuffer = None
        # List of sizes of buffer to be sent/recv in each direction
        self._buffer_size = []
        # Buffers for reception
        self._recvbuffer = None
        # domain dimension
        self._dim = self.topology.domain.dim

        self._setup_slices()

        # shape of numpy arrays to be updated.
        self.memshape = tuple(self.topology.mesh.local_resolution)
        # length of memory required to save one numpy array
        self._memoryblocksize = np.prod(self.memshape)
        # Number of numpy arrays that will be updated
        self.nb_elements = nb_elements

        # list of functions, _apply[d] corresponds to the synchronization
        # used in direction d, either
        #  - local (non-distributed) --> apply BC
        #  - distributed --> mpi comm
        self._apply = [None, ] * self._dim
        # distributed (mpi) directions
        self.exchange_dir = []
        if self.topology.size > 1:
            self.exchange_dir = [d for d in xrange(self._dim)
                                 if self.topology.cutdir[d]]
        for d in self.exchange_dir:
            self._apply[d] = self._apply_in_dir
        for d in self.topology.bc_dirs:
            self._apply[d] = self._apply_bc_in_dir

        if self.topology.size > 1:  # else no need to set buffers ...
            # Size computation below assumes that what we send in one
            # dir has the same size as what we receive from process in the
            # same dir ...
            # A temporary array used to calculate slices sizes
            temp = np.zeros(self.memshape, dtype=np.int8)

            for d in self.exchange_dir:
                buffsize = 0
                buffsize += temp[self._g_tonext[d]].size * self.nb_elements
                self._buffer_size.append(buffsize)
            self._recvbuffer = npw.zeros(max(self._buffer_size))
            self._sendbuffer = npw.zeros(max(self._buffer_size))

    def _setup_slices(self):
        """Compute slices used to describe what to send
        and receive in ghosts points.
        """
        # all points
        defslice = [slice(None, None, None)] * self._dim
        # no points
        nogh_slice = [slice(0)] * self._dim
        for d in xrange(self._dim):
            if self.ghosts[d] > 0:
                # get ghost points indices in current direction ...
                self._g_fromprevious.append(list(defslice))
                self._g_fromprevious[d][d] = slice(self.ghosts[d])
                self._g_fromnext.append(list(defslice))
                self._g_fromnext[d][d] = slice(-self.ghosts[d], None, None)
                self._g_toprevious.append(list(defslice))
                self._g_toprevious[d][d] = slice(self.ghosts[d],
                                                 2 * self.ghosts[d], None)
                self._g_tonext.append(list(defslice))
                self._g_tonext[d][d] = slice(-2 * self.ghosts[d],
                                             -self.ghosts[d])

                other_dim = self._other_dim(d)
                # ... and only compute points indices in other directions
                # i.e. 'corner' points are not taken into account.
                for d2 in other_dim:
                    self._g_fromprevious[d][d2] = slice(self.ghosts[d2],
                                                        -self.ghosts[d2])
                    self._g_fromnext[d][d2] = slice(self.ghosts[d2],
                                                    -self.ghosts[d2])
                    self._g_toprevious[d][d2] = slice(self.ghosts[d2],
                                                      -self.ghosts[d2])
                    self._g_tonext[d][d2] = slice(self.ghosts[d2],
                                                  -self.ghosts[d2])
            else:
                # empty lists of indices
                self._g_fromprevious.append(list(nogh_slice))
                self._g_fromnext.append(list(nogh_slice))
                self._g_toprevious.append(list(nogh_slice))
                self._g_tonext.append(list(nogh_slice))

        self._g_fromprevious = self._immutable(self._g_fromprevious)
        self._immutable(self._g_fromnext)
        self._immutable(self._g_toprevious)
        self._immutable(self._g_tonext)

    @staticmethod
    def _immutable(var):
        """Convert list of lists to immutable tuples
        """
        assert isinstance(var, list)
        for i in xrange(len(var)):
            var[i] = tuple(var[i])
        return tuple(var)

    def _other_dim(self, d):
        """utility to get all directions except d
        """
        return [x for x in xrange(self._dim) if x != d]

    def __call__(self, variables):
        return self.apply(variables)

    def apply_bc(self, variables):
        """Apply boundary conditions for non-distributed directions (only).

        Parameters
        ----------
        variables : a list of ndarrays

        Note that in distributed directions, BC are automatically set
        during ghosts update (apply).
        """
        assert (self.topology.domain.boundaries == PERIODIC).all(),\
            'Only implemented for periodic boundary conditions.'
        assert isinstance(variables, list)
        dirs = self.topology.bc_dirs
        for d in dirs:
            self._apply_bc_in_dir(variables, d, 0)

    def _apply_bc_in_dir(self, variables, d, i):
        """Apply periodic boundary condition in direction d."""
        assert (self.topology.domain.boundaries == PERIODIC).all(),\
            'Only implemented for periodic boundary conditions.'
        for v in variables:
            assert v.shape == self.memshape
            v[self._g_fromprevious[d]] = v[self._g_tonext[d]]
            v[self._g_fromnext[d]] = v[self._g_toprevious[d]]
        return i
    # @debug
    # def apply(self, variables):
    #     """Update ghosts points values:
    #     mpi communications and boundary conditions.

    #     Parameters
    #     ----------
    #     variables : list of numpy arrays
    #         arrays that must be synchronized.
    #     """
    #     assert isinstance(variables, list)
    #     # Apply boundary conditions for distributed directions ...
    #     i = 0
    #     for d in self.exchange_dir:
    #         self._apply_in_dir(variables, d, i)
    #         # update position in buffers list
    #         i += 1
    #     #  ... and for non-distributed directions.
    #     self.apply_bc(variables)

    @debug
    def apply(self, variables):
        """Update ghosts points values:
        mpi communications and boundary conditions.

        Parameters
        ----------
        variables : list of numpy arrays
            arrays that must be synchronized.
        """
        assert isinstance(variables, list)
        # Apply boundary conditions all directions ...
        i = 0
        for d in xrange(self._dim):
            i = self._apply[d](variables, d, i)

    def _apply_in_dir(self, variables, d, i):
        """Communicate ghosts values in direction d for neighbour
        in direction i of the topology

        Parameters
        ----------
        variables : list of numpy arrays
            arrays to be synchronized
        d : int
            direction of synchronization
        i : int
            counter, current position in buffers.
        """
        comm = self.topology.comm
        rank = self.topology.rank
        neighbours = self.topology.neighbours_tmp[d]
        # 1 - Fill in buffers
        # Loop through all variables that are distributed
        pos = 0
        nextpos = 0
        for v in variables:
            assert v.shape == self.memshape
            nextpos += v[self._g_tonext[d]].size
            self._sendbuffer[pos:nextpos] = v[self._g_tonext[d]].flat
            pos = nextpos

        # 2 - Send to next receive from previous
        dest_rk = neighbours[1]
        from_rk = neighbours[0]
        sendbuffer = self._sendbuffer[:self._buffer_size[i]]
        recvbuffer = self._recvbuffer[:self._buffer_size[i]]

        comm.Sendrecv([sendbuffer, hysop.core.mpi_REAL],
                      dest=dest_rk, sendtag=rank,
                      recvbuf=recvbuffer,
                      source=from_rk, recvtag=from_rk)

        # 3 - fill variables with recvbuffer and update sendbuffer
        # for next send
        pos = 0
        nextpos = 0
        for v in variables:
            nextpos += v[self._g_fromprevious[d]].size
            v[self._g_fromprevious[d]].flat = \
                self._recvbuffer[pos:nextpos]
            self._sendbuffer[pos:nextpos] = \
                v[self._g_toprevious[d]].flat
            pos = nextpos

        # 4 -Send to previous and receive from next
        dest_rk = neighbours[0]
        from_rk = neighbours[1]
        comm.Sendrecv([sendbuffer, hysop.core.mpi_REAL],
                      dest=dest_rk, sendtag=rank,
                      recvbuf=recvbuffer,
                      source=from_rk, recvtag=from_rk)
        # 5 - fill variables with recvbuffer
        pos = 0
        nextpos = 0
        for v in variables:
            nextpos += v[self._g_fromprevious[d]].size
            v[self._g_fromnext[d]].flat = \
                self._recvbuffer[pos:nextpos]
            pos = nextpos

        # update position in buffers list
        return i + 1


class UpdateGhostsFull(UpdateGhosts):
    """Ghost points synchronization for a list of numpy arrays

    Notes
    -----
    * This version differs from UpdateGhosts
    by computing also ghosts values
    in edges and corners of the domain.
    The directions are computed in reversed order.

    """

    def _other_dim(self, d):
        # Slices for other directions corresponding to directions not
        # yet exchanged : x > d. For directions x < d, the slices is a
        # full slice that includes ghosts. This assumes that directions
        # x < d have already been computed (by communications or local
        # exchanges)
        #                if d == 0:
        return [x for x in xrange(self._dim) if x > d]

    # @debug
    # def apply(self, variables):
    #     """Apply either mpi communications
    #     or local boundary conditions to fill ghosts.
    #     Loop over directions and switch among local BC or mpi comm.
    #     """
    #     assert isinstance(variables, list)
    #     local_bc_dir = self.topology.bc_dirs
    #     assert len(self.exchange_dir) + len(local_bc_dir) == self._dim

    #     i = 0
    #     for d in xrange(self._dim):
    #         if d in local_bc_dir:
    #             self._apply_bc_in_dir(variables, d)
    #         elif d in self.exchange_dir:
    #             self._apply_in_dir(variables, d, i)
    #             # update position in buffers list
    #             i += 1
    #             # End of loop through send/recv directions.
