"""Advection of a field.

See :ref:`advection` in user guide.
"""
from __future__ import print_function
from abc import ABCMeta, abstractmethod
from hysop.constants import debug, DirectionLabels, ZDIR
from hysop import __SCALES_ENABLED__, Field, __GPU_ENABLED__
from hysop.operator.computational import Computational
from hysop.methods import TimeIntegrator, Interpolation,\
    Remesh, Support, Splitting, MultiScale
from hysop.numerics.remeshing import Remeshing, L2_1
from hysop.operator.continuous import opsetup, opapply
import hysop.default_methods as default
from hysop.tools.parameters import Discretization
from hysop.topology.cartesian_topology import CartesianTopology
from hysop.tools.numpywrappers import npw
from hysop.problem.simulation import O2, SplittingParameters
from hysop.operator.discrete.particle_advection import ParticleAdvection
import numpy as np
if __GPU_ENABLED__:
    from hysop.backend.device.opencl.gpu_particle_advection import GPUParticleAdvection
    from hysop.backend.device.opencl.multi_gpu_particle_advection import MultiGPUParticleAdvection
if __SCALES_ENABLED__:
    from hysop.f2hysop import scales2py as scales


class AdvectionBase(Computational):
    """Abstract interface to advection operators
    """
    __metaclass__ = ABCMeta

    @debug
    def __init__(self, velocity, advected_fields=None,
                 discretization_fields=None, **kwds):
        """
        Parameters
        ----------
        velocity : :class:`hysop.field.continuous.Field`
            the velocity field
        advected_field: a list of :class:`hysop.field.continuous.Field`,
         optional
            fields to be advected
        discretization_fields : :class:`hysop.topology.topology.CartesianTopology`
         or :class:`tools.parameters.Discretization`
            Defined the data/mpi distribution for advected_fields.
            Default = same as velocity
        kwds : base class arguments.

        Notes
        -----
        * one single discretization for all advected fields
        * velocity discretization may be different
        * to limit advection to one or two directions, set
          directions=[0, 2] (advection in x and z dirs).

        """
        # --- check inputs for fields (advected and velocity) ---
        msg = 'discretization argument is required in Advection.'
        assert 'discretization' in kwds, msg
        msg = 'Wrong input for advection operator:'
        msg += ' "variables" attribute is not allowed. Check manual.'
        assert 'variables' not in kwds, msg
        # advected_fields :
        # == field or [field1, field2, ...] ==> same discretization for all,
        #  equal to velocity discretization if discretization_fields is None
        self.advected_fields = []
        if advected_fields is not None:
            if not isinstance(advected_fields, list):
                assert isinstance(advected_fields, Field)
                advected_fields = [advected_fields]
            self.advected_fields = advected_fields
        # velocity field
        if discretization_fields is not None:
            dfields = discretization_fields
        else:
            dfields = kwds['discretization']
        variables = {velocity: kwds['discretization']}
        for field in advected_fields:
            variables[field] = dfields
        kwds.pop('discretization')
        super(AdvectionBase, self).__init__(variables=variables, **kwds)
        # velocity field
        self.velocity = velocity
        assert velocity.nb_components == self.domain.dim

        # --- Set default method, if required ---
        if self.method is None:
            self.method = default.ADVECTION

        # --- Set name ---
        vars_str = "_("
        for vv in self.advected_fields:
            vars_str += vv.name + ","
        self.name += vars_str[0:-1] + ')'

        self.config = {}
        self.input = [var for var in self.variables]
        self.output = [var for var in self.variables
                       if var is not self.velocity]

    def get_work_properties(self):
        return {'rwork': None, 'iwork': None}

    @opsetup
    def setup(self, rwork=None, iwork=None):
        if self._is_uptodate:
            return
        # select discretization of the advected fields
        advected_discrete_fields = [self.discrete_fields[v]
                                    for v in self.variables
                                    if v is not self.velocity]
        toporef = advected_discrete_fields[0].topology
        msg = 'All advected fields must have the same topology.'
        for f in advected_discrete_fields:
            assert f.topology == toporef, msg

        # call specific setup (see derived class)
        self._setup(advected_discrete_fields, rwork, iwork)
        self._is_uptodate = True

    def advected_fields_topology(self):
        """Returns the topology used for advected fields
        """
        msg = 'Operator must be discretized'
        assert self._is_discretized, msg
        return self.discrete_fields[self.advected_fields[0]].topology

    def velocity_topology(self):
        """Returns the topology used for velocity
        """
        msg = 'Operator must be discretized'
        assert self._is_discretized, msg
        return self.discrete_fields[self.velocity].topology

    @abstractmethod
    def _setup(self, advected_discrete_fields, rwork=None, iwork=None):
        """Setup specific to advection implementation (see derived class)
        """


class Advection(AdvectionBase):
    """Python or GPU advection
    """
    @debug
    def __init__(self, directions=None, **kwds):
        """
        Parameters
        ----------
        directions : list of int, optional
            direction(s) in which advection must be performed. Default = all.
        kwds : base class arguments.

        """
        super(Advection, self).__init__(**kwds)
        self.extra_args = {} ### TO BE REVIEWED ###
        if directions is None:
            directions = [i for i in xrange(self.domain.dim)]
        assert isinstance(directions, list)
        self.directions = directions
        assert TimeIntegrator in self.method.keys()
        assert Interpolation in self.method.keys()
        assert Remesh in self.method.keys()
        if Splitting not in self.method.keys():
            self.method[Splitting] = O2
        self.discrete_op = [None] * self.domain.dim

        # Fields on particles
        self.particle_fields = None

        # Positions of the particles
        self.particle_positions = None

        # number of components of the rhs (time integration)
        self._rhs_size = 1

        # check if advection is done on gpu
        if Support in self.method:
            self._gpu_advection = self.method[Support].find('gpu') >= 0
        else:
            self._gpu_advection = False
        self.splitting = None
        self._old_dir = 0

    def discretize(self):
        """Discretization (create topologies and discretize fields)
        """
        if self._is_discretized:
            return
        cutdir = [False] * self.domain.dim
        cutdir[-1] = True
        # call standard discretization (from computational base class)
        self._standard_discretize(cutdir=cutdir)
        # check if velocity and fields are defined with the same resolution
        if self._single_topo:
            self.method[MultiScale] = None
        else:
            # set a default value for interpolation method
            if MultiScale not in self.method or\
               self.method[MultiScale] is None:
                self.method[MultiScale] = L2_1
            mscale = self.method[MultiScale]
            min_ghosts = mscale.ghosts_layer_size
            # get ghost layer size used for velocity field
            ghosts_v = self.variables[self.velocity].ghosts()
            # and check if it fits with interpolation method
            msg = 'Ghost layer required for velocity. Size min = '
            msg += str(min_ghosts) + " (" + str(ghosts_v) + " given)"
            assert (ghosts_v >= min_ghosts).all(), msg
        self._is_discretized = True

    def get_work_properties(self):
        assert self._is_discretized
        # Shape of reference comes from fields, not from velocity
        advected_discrete_fields = [self.discrete_fields[v]
                                    for v in self.variables
                                    if v is not self.velocity]
        topo = advected_discrete_fields[0].topology
        # Find number and shape of required work arrays
        if not self._gpu_advection:
            # -- pure python advection --
            #  work array shape depends on the time integrator
            #  interpolation scheme and remeshing scheme
            ti_work = self.method[TimeIntegrator].get_work_properties(
                self._rhs_size, topo)
            ti_rwork_length = len(ti_work['rwork'])
            iw_prop = self.method[Interpolation].get_work_properties(topo)
            rw_prop = Remeshing.get_work_properties(topo)
            interp_iwork_length = len(iw_prop['iwork'])
            interp_rwork_length = len(iw_prop['rwork'])
            remesh_iwork_length = len(rw_prop['iwork'])
            remesh_rwork_length = len(rw_prop['rwork'])
            iwork_length = max(interp_iwork_length, remesh_iwork_length)
            rwork_length = max(ti_rwork_length + interp_rwork_length,
                               remesh_rwork_length)

        else:
            # -- GPU advection --
            # no work array
            iwork_length, rwork_length = 0, 0

        # buffers for fields on particles
        rwork_length += np.sum([f.nb_components for f in self.advected_fields])
        if not self._gpu_advection or \
           self.method[Support].find('gpu_2k') >= 0:
            rwork_length += 1  # work array for positions
        memsize = np.prod(topo.mesh.local_resolution)
        return {'rwork': [(memsize,)] * rwork_length,
                'iwork': [(memsize,)] * iwork_length}

    def _setup(self, advected_discrete_fields, rwork=None, iwork=None):
        for i in self.directions:
            ref_shape = advected_discrete_fields[0].topology.shape[i]
            if self._gpu_advection:
                if ref_shape == 1:
                    DA = GPUParticleAdvection
                else:
                    DA = MultiGPUParticleAdvection
            else:
                DA = ParticleAdvection

            self.discrete_op[i] = DA(
                velocity=self.discrete_fields[self.velocity],
                fields_on_grid=advected_discrete_fields,
                direction=i, method=self.method,
                rwork=rwork, iwork=iwork)
                #**self.extra_args)

            if i == 0:
                # work arrays can be shared between directions.
                rwork = self.discrete_op[i]._rwork
                iwork = self.discrete_op[i]._iwork

        # set splitting parameters (depends on method)
        self.splitting = self.method[Splitting](self.domain.dim)
        assert isinstance(self.splitting, SplittingParameters)

        # configure gpu
        if self._gpu_advection:
            self._configure_gpu()

    def _configure_gpu(self):
        """Setup for gpu related things (Memory ...)
        """
        splitting_nbSteps = len(self.splitting)
        for d in xrange(self.domain.dim):
            dOp = self.discrete_op[d]
            assert len(dOp.exec_list) == splitting_nbSteps, \
                "Discrete operator execution " + \
                "list and splitting steps sizes must be equal " + \
                str(len(dOp.exec_list)) + " != " + \
                str(splitting_nbSteps)
        s = ""
        device_id = self.discrete_op[0].cl_env._device_id
        gpu_comm = self.discrete_op[0].cl_env.gpu_comm
        gpu_rank = gpu_comm.Get_rank()
        if gpu_rank == 0:
            s += "=== OpenCL buffers allocated"
            s += " on Device:{0} ===\n".format(device_id)
            s += "Global memory used:\n"
        total_gmem = 0
        for d in xrange(self.domain.dim):
            g_mem_d = 0
            # allocate all variables in advec_dir
            for df in self.discrete_op[d].variables:
                if not df.gpu_allocated:
                    df.allocate()
                    g_mem_df = gpu_comm.allreduce(df.mem_size)
                    g_mem_d += g_mem_df
            if gpu_rank == 0:
                s += " Advection" + DirectionLabels[d] + ": {0:9d}".format(g_mem_d)
                s += "Bytes ({0:5d} MB)\n".format(g_mem_d / (1024 ** 2))
            total_gmem += g_mem_d
        if gpu_rank == 0:
            s += " Total      : {0:9d}".format(total_gmem)
            s += "Bytes ({0:5d} MB)\n".format(total_gmem / (1024 ** 2))
            s += "Local memory used:\n"
        total_lmem = 0
        for d in xrange(self.domain.dim):
            l_mem_d = gpu_comm.allreduce(
                self.discrete_op[d].size_local_alloc)
            if gpu_rank == 0:
                s += " Advection" + DirectionLabels[d] + ": {0:9d}".format(l_mem_d)
                s += "Bytes ({0:5d} MB)\n".format(l_mem_d / (1024 ** 2))
            total_lmem += l_mem_d
        if gpu_rank == 0:
            s += " Total      : {0:9d}".format(total_lmem) + "Bytes"
            print(s)

    @debug
    @opapply
    def apply(self, simulation=None):
        """Redefinition of apply for advection --> dimensional splitting.

        Parameters
        ----------
        simulation : `:class::~hysop.problem.simulation.Simulation`

        """
        assert simulation is not None
        for split_id, split in enumerate(self.splitting()):
            simulation.set_split(split_id, split)
            self.discrete_op[simulation.current_dir].apply(simulation)
            #, split[1], split_id, self._old_dir)
            simulation.next_split()

    @debug
    def finalize(self):
        """Memory cleaning.
        """
        for i in self.directions:
            self.discrete_op[i].finalize()

    def get_profiling_info(self):
        if self._is_uptodate:
            for d in self.directions:
                self.profiler += self.discrete_op[d].profiler

    def __str__(self):
        """
        Common printings for operators
        """
        short_name = str(self.__class__).rpartition('.')[-1][0:-2]
        for i in self.directions:
            if self.discrete_op[i] is not None:
                s = str(self.discrete_op[i])
            else:
                s = short_name + " operator. Not discretised."
        return s + "\n"


class ScalesAdvection(AdvectionBase):
    """Advection based on fortran (scales) interface
    """
    @debug
    def __init__(self, cutdir=None, **kwds):
        """
        Parameters
        ----------
        cutdir : list of bool
            cutdir[d] = True to distribute data in direction d.

        kwds : base class arguments.

        Notes
        -----
        * In scales cutdir[0] must be False, i.e. data are not
        distributed in the first dir, which corresponds to contiguous
        memory layout (fortran).
        * Scales assumes that all subdomains have the same
          local resolution.
        * No ghosts points allowed in scales (at the moment)

        """
        super(ScalesAdvection, self).__init__(**kwds)
        assert Remesh in self.method.keys()
        msg = 'Scales is not available for your configuration.'
        assert __SCALES_ENABLED__, msg
        msg = 'Scales Advection is only implemented in 3D.'
        assert self.domain.dim == 3, msg
        # Default splitting = Strang
        if Splitting not in self.method.keys():
            self.method[Splitting] = 'strang'
        if cutdir is None:
            cutdir = [False, ] * self.domain.dim
            cutdir[-1] = True
        self.cutdir = cutdir

    def scales_parameters(self):
        """
        Return the name of the particular method used in scales
        and the type of splitting.
        """
        order = None
        for o in ['p_O2', 'p_O4', 'p_L2',
                  'p_M4', 'p_M6', 'p_M8',
                  'p_44', 'p_64', 'p_66', 'p_84']:
            if self.method[Remesh].find(o) >= 0:
                order = o
        if order is None:
            print('Unknown advection method, turn to default (p_M6).')
            order = 'p_M6'

        # - Extract splitting form self.method (default strang) -
        splitting = 'strang'
        for s in ['classic', 'strang', 'particle']:
            if self.method[Splitting].find(s) >= 0:
                splitting = s

        return order, splitting

    def discretize(self):
        """
        Discretization (create topologies and discretize fields)
        Available methods :
        - 'scales' : SCALES fortran routines (3d only, list of vector
        and/or scalar)
        - 'gpu' : OpenCL kernels (2d and 3d, single field, scalar or vector)
        - other : Pure python (2d and 3d, list of vector and/or scalar)
        """
        if self._is_discretized:
            return
        # Check if topos need to be created
        build_topos = self._check_variables()
        order, splitting = self.scales_parameters()

        # Scales, single resolution
        if self._single_topo:
            if build_topos:
                # In that case, self._discretization must be
                # a Discretization object, used for all fields.
                # We use it to initialize scales solver
                topo = self._create_scales_topo(self._discretization,
                                                order, splitting)
                for v in self.variables:
                    self.variables[v] = topo
            else:
                # In that case, self._discretization must be
                # a CartesianTopology object, used for all fields.
                # We use it to initialize scales solver
                assert isinstance(self._discretization, CartesianTopology)
                topo = self._discretization
                msg = 'input topology is not compliant with scales.'
                #assert topo.dimension == 1, msg
                msg = 'Ghosts points not yet implemented for scales operators.'
                assert (topo.mesh.discretization.ghosts == 0).all(), msg

                nbcells = topo.mesh.discretization.resolution - 1
                topodims = topo.shape
                scalesres, global_start = \
                    scales.init_advection_solver(nbcells,
                                                 self.domain.length,
                                                 npw.asintegerarray(topodims),
                                                 self.mpi_params.comm.py2f(),
                                                 order=order,
                                                 dim_split=splitting)

                assert (topo.shape == topodims).all()
                assert (topo.mesh.local_resolution == scalesres).all()
                assert (topo.mesh.start() == global_start).all()

            msg = 'Scales Advection not yet implemented with ghosts points.'
            assert (topo.ghosts() == 0).all(), msg

        # Scales, multi-resolution
        else:
            if build_topos[self.velocity]:
                # Resolution used for velocity
                v_resol = self.variables[self.velocity].resolution - 1

            else:
                topo = self.variables[self.velocity]
                v_resol = topo.mesh.discretization.resolution

            vbuild = [v for v in self.variables if build_topos[v]]
            for v in vbuild:
                self.variables[v] = self._create_scales_topo(
                    self.variables[v], order, splitting)

            topo = self.variables.values()[0]
            self._check_scales_topo(topo, order, splitting)

            # Init multiscale in scales
            scales.init_multiscale(v_resol[0], v_resol[1], v_resol[2],
                                   self.method[MultiScale])

        # All topos are built, we can discretize fields.
        self._discretize_vars()

        advected_discrete_fields = [self.discrete_fields[f]
                                    for f in self.advected_fields]
        toporef = advected_discrete_fields[0].topology
        msg = 'All advected fields must have the same topology.'
        for f in advected_discrete_fields:
            assert f.topology == toporef, msg

        if self._single_topo:
            self.method[MultiScale] = None

    def _create_scales_topo(self, d3d, order, splitting):
        """set specific MPI layout for scales.
        """
        comm = self.mpi_params.comm
        topodims = [1, 1, comm.Get_size()]
        msg = 'Wrong type for parameter discretization (at init).'
        msg += str(self._discretization)
        assert isinstance(d3d, Discretization), msg
        nbcells = d3d.resolution - 1
        scalesres, global_start = \
            scales.init_advection_solver(nbcells,
                                         self.domain.length,
                                         npw.asintegerarray(topodims),
                                         comm.py2f(),
                                         order=order,
                                         dim_split=splitting)
        # Create the topo (plane, cut through ZDIR)
        return self.domain.create_plane_topology_from_mesh(
            global_start=global_start, localres=scalesres,
            discretization=d3d, cdir=ZDIR)

    def _check_scales_topo(self, toporef, order, splitting):
        """Check if input topo fits with scales requirements
        """
        # In that case, self._discretization must be
        # a CartesianTopology object, used for all fields.
        # We use it to initialize scales solver
        comm = self.mpi_params.comm
        #topodims = [1, 1, comm.Get_size()]
        nbcells = toporef.mesh.discretization.resolution - 1

        scalesres, global_start = \
            scales.init_advection_solver(nbcells, self.domain.length,
                                         npw.asintegerarray(toporef.shape),
                                         comm.py2f(),
                                         order=order, dim_split=splitting)
        for v in self.variables:
            topo = self.variables[v]
            assert isinstance(topo, CartesianTopology), str(topo)
            #assert (topo.shape == topodims).all(), \
            #    str(topo.shape) + ' != ' + str(topodims)
            assert not self._single_topo or \
                (topo.mesh.local_resolution == scalesres).all(), \
                str(topo.mesh.local_resolution) + ' != ' + str(scalesres)
            assert not self._single_topo or \
                (topo.mesh.start() == global_start).all(), \
                str(topo.mesh.start()) + ' != ' + str(global_start)

    def _setup(self, advected_discrete_fields, rwork=None, iwork=None):
        # Check resolutions to set multiscale case, if required.
        if not self._single_topo and MultiScale not in self.method:
            self.method[MultiScale] = L2_1
        # - Create the discrete_op from the
        # list of discrete fields -
        from hysop.operator.discrete.scales_advection import \
            ScalesAdvection as SD
        self.discrete_op = SD(
            self.discrete_fields[self.velocity],
            advected_discrete_fields, method=self.method,
            rwork=rwork, iwork=iwork,
            **self.config)
