# -*- coding: utf-8 -*-
"""Operator to kill the vorticity at the outlet boundary
(i.e. removal of the periodic BC in the flow direction
by vorticity absorption in order to set the far field
velocity to u_inf at the inlet)
"""

from hysop.constants import debug, np
from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.fields.variable_parameter import VariableParameter
from hysop.tools.profiler import profile
from hysop.tools.numpywrappers import npw
from hysop.constants import YDIR, ZDIR
from hysop.tools.misc import WorkSpaceTools


class AbsorptionBC(DiscreteOperator):
    """
    The periodic boundary condition is modified at the outlet
    in the flow direction in order to discard
    in the downstream region the eddies coming
    periodically from the outlet.
    The vorticity absorption conserves div(omega)=0.
    The far field velocity is set to u_inf at the inlet.
    """

    @debug
    def __init__(self, velocity, vorticity, req_flowrate,
                 absorption_box, filter_func=None, **kwds):
        """
        Parameters
        ----------
        velocity, vorticity : :class:`~hysop.fields.discrete_field.DiscreteField`
        req_flowrate : double
            required value for the flow rate
        absorption_box : :class:`~hysop.subsets.SubBox`
            a box representing the area where filter is applied.
        filter_func: list of python functions, optional
            functions used to compute the filter and its differential.
        **kwds : extra parameters for base class


        Notes
        -----
        * if set, filter_func[0] and filter_func[1] must be python function
        returning a numpy array. For example to apply a sine inside
        the absorption area use :

        .. code::

            def func(x):
                return np.sin(x)

        """
        assert 'variables' not in kwds, 'variables parameter is useless.'
        # velocity discrete field
        self.velocity = velocity
        # vorticity discrete field
        self.vorticity = vorticity
        self.absorption_box = absorption_box
        super(AbsorptionBC, self).__init__(
            variables=[velocity, vorticity], **kwds)
        # If 2D problem, vorticity must be a scalar
        if self._dim == 2:
            assert self.vorticity.nb_components == 1
        assert (self._dim > 2),\
            "Wrong problem dimension: only 3D cases are implemented."
        topo = self.vorticity.topology
        xcoords = topo.mesh.coords[0]
        self.input = self.variables
        self.output = [self.vorticity]
        # Expected value for the flow rate through self.surfRef
        msg = 'Wrong type or length for input req_flowrate.'
        if not isinstance(req_flowrate, VariableParameter):
            self.req_flowrate = VariableParameter(
                data=np.asarray(req_flowrate))
            assert np.asarray(req_flowrate).size == self._dim, msg
        else:
            self.req_flowrate = req_flowrate

        t_dir = [1, 2]
        dsurf = npw.prod(self.absorption_box.real_length[topo][t_dir])
        self._inv_ds = 1. / dsurf
        if self.absorption_box.on_proc[topo]:
            ind = self.absorption_box.ind[topo][0]

            if filter_func is None:
                self.absorption_filter = None
                self.diff_filter = None
                self._set_default_filter(xcoords[ind[0]])
            else:
                assert isinstance(filter_func, list)
                self.absorption_filter = filter_func[0](xcoords[ind[0]])
                self.diff_filter = filter_func[1](xcoords[ind[0]])

    def _set_default_filter(self, x):
        """Default values for the filter in the absorption box
        """
        xb = x[0]
        xe = x[-1]
        xc = xb + (xe - xb) / 2.0
        eps = 10.
        form = np.tanh(eps * (x - xc))
        self.absorption_filter = form - np.tanh(eps * (xe - xc))
        coeff = 1.0 / (np.tanh(eps * (xb - xc)) - np.tanh(eps * (xe - xc)))
        self.absorption_filter[...] *= coeff
        self.diff_filter = eps * (1.0 - form ** 2)
        self.diff_filter *= coeff

    def _set_work_arrays(self, rwork=None, iwork=None):

        # Reference shape comes from the box at the end of the domain
        subshape = tuple(
            self.absorption_box.mesh[self.vorticity.topology].resolution)
        self._rwork = WorkSpaceTools.check_work_array(1, subshape, rwork)

    @debug
    @profile
    def apply(self, simulation=None):
        if not self.absorption_box.on_proc[self.vorticity.topology]:
            return
        # the required flowrate value is updated (depending on time)
        self.req_flowrate.update(simulation)
        # \warning : the flow rate value is divided by area of input surf.
        req_flowrate_val = self.req_flowrate.data * self._inv_ds

        ind = self.absorption_box.ind[self.vorticity.topology][0]
        # 1 - filter * periodic vorticity:
        for d in xrange(self.vorticity.nb_components):
            np.multiply(self.vorticity[d][ind], self.absorption_filter,
                        self.vorticity[d][ind])

        # 2 - nabla filter X periodic velocity + nabla(1-filter) X uinf
        # uinf - vz
        np.subtract(req_flowrate_val, self.velocity[ZDIR][ind],
                    self._rwork[0])
        # wk = filter' * (uinf - vz )
        np.multiply(self.diff_filter, self._rwork[0], self._rwork[0])
        # add the result into vorticity
        np.add(self.vorticity.data[YDIR][ind], self._rwork[0],
               self.vorticity.data[YDIR][ind])
        # and now, w_z ...
        # vy - uinf
        np.subtract(self.velocity[YDIR][ind], req_flowrate_val,
                    self._rwork[0])
        # wk = filter' * (uinf - vy )
        np.multiply(self.diff_filter, self._rwork[0], self._rwork[0])
        # add the result into vorticity
        np.add(self.vorticity.data[ZDIR][ind], self._rwork[0],
               self.vorticity.data[ZDIR][ind])
