# -*- coding: utf-8 -*-
"""Discrete operator to Compute the time step according to the flow fields.
"""

from hysop.constants import debug
from hysop.methods import TimeIntegrator, SpaceDiscretization
from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.numerics.differential_operations import MaxDiagGradV, \
    DiagAndStretch, StrainCriteria, StretchLike, StrainAndStretch
from hysop.tools.numpywrappers import npw
from hysop.numerics.update_ghosts import UpdateGhosts
from hysop.core.mpi import MPI
from hysop.constants import np, hysop.core.mpi_REAL
from hysop.tools.profiler import profile
from hysop.problem.simulation import Simulation
from hysop.tools.misc import WorkSpaceTools


class AdaptiveTimeStepD(DiscreteOperator):
    """
    The adaptative Time Step is computed as:

    dt_adapt = min (dt_advection, dt_stretching, dt_cfl)
    """

    authorized_criteria = ['vort', 'gradU', 'stretch', 'cfl', 'deform']

    @debug
    def __init__(self, velocity, vorticity, simulation,
                 lcfl=0.125, cfl=0.5, criteria=None,
                 time_range=None, maxdt=9999., **kwds):
        """
        Parameters
        ----------
        velocity, vorticity : :class:`~hysop.fields.discrete_field.DiscreteFields`
            discrete fields used to update the time steps. Read only.
        simulation : :class:`~hysop.problem.simulation.Simulation`
        lcfl : double, optional
            the lagrangian CFL coefficient used for advection stability
        cfl : double, optional
            CFL coefficient
        criteria : list of strings, optional
            name of the criteria used to compute the new time step.
            See notes below.
        time_range : list of two integers, optional
            use to define a 'window' in which the current operator is applied.
            time_range = [start, end], outside this range, the operator
            has no effect. Start/end are iteration numbers.
            Default = [2, end of simu]
        maxdt : double, optional
            maximum value allowed for time step. Default = 9999.
        kwds : arguments passed to base class.

        Notes
        -----
        * This operator has no effect on input variables.
        * Authorized criteria are:
            * criteria for the advection scheme, one of :
                1. 'vort' (default)
                2. 'gradU'
                3. 'strain'
            * criteria for the stretching scheme :
                4. 'stretch'
            * cfl-like :
                5. 'cfl'
        * Computes a 'diagnostics' vector :
          diagnostics = (time, time_step, c1, c2, ...)
          ci = time step computed from the criterium number i in the list
          above.
        """
        # velocity discrete field
        self.velocity = velocity
        # vorticity discrete field
        self.vorticity = vorticity
        # adaptative time step variable
        assert isinstance(simulation, Simulation)
        self.simulation = simulation
        # Build the user required function list
        self._used_functions = {}
        # local buffer used to compute max values.
        self._result = None
        # position in result (only needed when stretch criterium is used)
        self._pos = 0
        # name of (optional) differential operator required to compute
        # the criteria.
        self._diffop_name = self._init_differential_operator(criteria)
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(AdaptiveTimeStepD, self).__init__(
            variables=[velocity, vorticity], **kwds)

        self.input = self.variables
        self.output = []
        # Courant Fredrich Levy coefficient
        self.cfl = cfl
        # Lagrangian CFL coefficient
        self.lcfl = lcfl
        # Max. timestep
        self.maxdt = maxdt

        # Time range
        if time_range is None:
            time_range = [2, np.infty]
        self.time_range = time_range

        # local buffer :
        # [time, dt, d1, d2, d3, d4, d5]
        # for d1...d5 see computation details in apply.
        self.diagnostics = npw.zeros((7))
        # list of indices of diagnostics of interest.
        self._diag_ind = []
        # True if ghost points synchro is required
        self._needs_synchro = False
        # The (optional) differential operation used to
        # compute the criterium
        self._diff_op = None
        # Coefficient used to compute streching stability criterium
        self.coeff_stretch = self.method[TimeIntegrator].stability_coeff()
        # Init functions ...
        if self._diffop_name is not None:
            self._diff_op = self._diffop_name(
                topo=self.velocity.topology,
                method=self.method[SpaceDiscretization],
                work=self._rwork)
            self._needs_synchro = True
            # Ghost synchronisation operator.
            if 'vort' in criteria and len(criteria) == 1:
                self._synchronize = UpdateGhosts(self.vorticity.topology,
                                                 self.vorticity.nb_components)
                self._synchr_buff = self.vorticity.data
            elif 'vort' in criteria:
                self._synchronize = UpdateGhosts(
                    self.velocity.topology,
                    self.vorticity.nb_components + self.velocity.nb_components)
                self._synchr_buff = self.velocity.data + self.vorticity.data

            else:
                self._synchronize = UpdateGhosts(
                    self.velocity.topology,
                    self.velocity.nb_components)
                self._synchr_buff = self.velocity.data

    def _set_work_arrays(self, rwork=None, iwork=None):
        """Allocate local work spaces (if needed)
        """
        if self._diffop_name is not None:
            topo = self.velocity.topology
            wkp = self._diffop_name.get_work_properties(topo)
            lwork = len(wkp['rwork'])
            subshape = wkp['rwork'][0]
            self._rwork = WorkSpaceTools.check_work_array(lwork,
                                                          subshape, rwork)

    def _init_differential_operator(self, criteria):
        """Select the required differential operator (if any)
        depending on the chosen criteria

        Notes:
        * required to check/allocate internal work arrays
        * required to connect the functions needed to compute the criteria
        * MUST be called before parent class init
        """
        # Dictionnary of the available diagnostics functions
        # all_functions[name] =
        # (Index in self.diagnostics, function)
        all_functions = {
            'vort': (2, self._compute_vort),
            'gradU': (3, self._compute_grad_u),
            'strain': (4, self._compute_strain),
            'stretch': (5, self._compute_stretch),
            'cfl': (6, self._compute_cfl)
            }

        # select functions of interest, from input criteria list.
        self._used_functions = {all_functions[k][0]: all_functions[k][1]
                                for k in criteria if k in all_functions}
        diffop = None
        dimension = self.velocity.domain.dim
        if 'gradU' in criteria:
            diffop = MaxDiagGradV
            self._result = npw.zeros(dimension)
        if 'stretch' in criteria:
            diffop = StretchLike
            self._result = npw.zeros(dimension)
            self._pos = 0
        if 'strain' in criteria:
            diffop = StrainCriteria
            self._result = npw.zeros(dimension)
        if 'stretch' in criteria and 'gradU' in criteria:
            diffop = DiagAndStretch
            self._result = npw.zeros(2 * dimension)
            self._pos = dimension
        if 'stretch' in criteria and 'strain' in criteria:
            diffop = StrainAndStretch
            self._result = npw.zeros(2 * dimension)
            self._pos = dimension
        return diffop

    def _compute_grad_u(self):
        """Compute criterium ...
        """
        self._result = self._diff_op(self.velocity.data, self._result)
        return self.lcfl / max(self._result[:self._dim])

    def _compute_stretch(self):
        """Compute criterium ...
        """
        self._result = self._diff_op(self.velocity.data, self._result)
        return self.coeff_stretch / max(
            self._result[self._pos:self._pos + self._dim])

    def _compute_cfl(self):
        """maxima of velocity : needed for CFL based time step
        """
        coef = self.cfl * self.velocity.topology.mesh.space_step[0]
        return coef / np.max([np.max(np.abs(v_c))
                              for v_c in self.velocity.data])

    def _compute_vort(self):
        """maxima of vorticity :
        needed for advection stability condition
        """
        return self.lcfl / np.max([np.max(np.abs(w_c))
                                   for w_c in self.vorticity.data])

    def _compute_strain(self):
        """maxima of strain tensor:
        needed for advection stability condition
        """
        self._result = self._diff_op(self.velocity.data, self._result)
        return self.lcfl / np.max(self._result[:self._dim])

    @debug
    @profile
    def apply(self, simulation=None):
        if simulation is not None:
            assert self.simulation is simulation
        ind = self.velocity.topology.mesh.compute_index
        # current time and iteration number
        time = self.simulation.time
        iteration = self.simulation.current_iteration
        # upper value allowed for iterations
        nmax = min(self.simulation.max_iter, self.time_range[1])
        buff = npw.zeros(7)
        buff[0] = time
        if iteration >= self.time_range[0] and iteration <= nmax:
            # Apply each function listed in used_functions
            if self._needs_synchro:
                # Synchronize ghost points of velocity
                self._synchronize(self._synchr_buff)

            for func in self._used_functions:
                # func: (Index in self.diagnostics, function, is gradU needed)
                buff[func] = self._used_functions[func]()

            # mpi reduction
            self.velocity.topology.comm.Allreduce(
                sendbuf=[buff, 7, hysop.core.mpi_REAL],
                recvbuf=[self.diagnostics, 7, hysop.core.mpi_REAL],
                op=MPI.MAX)
            ind = self._used_functions.keys()
            time_step = np.min(list(self.diagnostics[ind]) +
                               [self.maxdt])
            self.diagnostics[0] = time
            self.diagnostics[1] = time_step

            if self._writer is not None and self._writer.do_write(iteration):
                self._writer.buffer[0, :] = self.diagnostics
                self._writer.write()

            # Update simulation time step with the new dt
            self.simulation.update_time_step(time_step)
            # Warning this update is done only for the current MPI task!
            # See wait function in base class.
