# -*- coding: utf-8 -*-
"""Discrete operator to compute Energy and Enstrophy
"""
from hysop.constants import debug
from hysop.tools.profiler import profile
from hysop.tools.numpywrappers import npw
from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.tools.misc import WorkSpaceTools


class EnergyEnstrophy(DiscreteOperator):
    """Discretization of the energy/enstrophy computation process.
    """
    def __init__(self, velocity, vorticity, is_normalized=True, **kwds):
        """
        Parameters
        ----------
        velocity : :class:`~hysop.operator.field.discrete.DiscreteField`
            velocity discrete fields
        vorticity : :class:`~hysop.operator.field.discrete.DiscreteField`
            vorticity discrete fields
        isNormalized : boolean
            true if enstrophy and energy values have to be normalized
            by the domain lengths.
        """
        # velocity field
        self.velocity = velocity
        # vorticity field
        self.vorticity = vorticity

        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(EnergyEnstrophy, self).__init__(variables=[velocity, vorticity],
                                              **kwds)
        # Coeffs for integration
        self.coeff = {}
        # Global energy
        self.energy = 0.0
        # Global enstrophy
        self.enstrophy = 0.0
        topo_w = self.vorticity.topology
        topo_v = self.velocity.topology
        space_step = topo_w.mesh.space_step
        length = topo_w.domain.length
        # remark topo_w.domain and topo_v.domain
        # must be the same, no need to topo_v...length.
        self.coeff['Enstrophy'] = npw.prod(space_step)
        space_step = topo_v.mesh.space_step
        self.coeff['Energy'] = 0.5 * npw.prod(space_step)
        if is_normalized:
            normalization = 1. / npw.prod(length)
            self.coeff['Enstrophy'] *= normalization
            self.coeff['Energy'] *= normalization

    def _set_work_arrays(self, rwork=None, iwork=None):

        v_ind = self.velocity.topology.mesh.compute_index
        w_ind = self.vorticity.topology.mesh.compute_index
        size_v = self.velocity.data[0][v_ind].size
        size_w = self.vorticity.data[0][w_ind].size
        size_work = max(size_w, size_v)
        lwork = 1
        self._rwork = WorkSpaceTools.check_work_array(lwork, size_work, rwork)

    @debug
    @profile
    def apply(self, simulation=None):
        assert simulation is not None, \
            "Missing simulation value for computation."

        # --- Kinetic energy computation ---
        vd = self.velocity
        # get the list of computation points (no ghosts)
        nbc = vd.nb_components
        ic = self.velocity.topology.mesh.compute_index
        # Integrate (locally) velocity ** 2
        local_energy = 0.
        size_v = vd[0][ic].size
        for i in xrange(nbc):
            self._rwork[0][:size_v] = (vd[i][ic] ** 2).flat
            local_energy += npw.real_sum(self._rwork[0][:size_v])

        # --- Enstrophy computation ---
        wd = self.vorticity
        nbc = wd.nb_components
        ic = self.vorticity.topology.mesh.compute_index
        size_w = wd[0][ic].size
        # Integrate (locally) vorticity ** 2
        local_enstrophy = 0.
        for i in xrange(nbc):
            self._rwork[0][:size_w] = (wd[i][ic] ** 2).flat
            local_enstrophy += npw.real_sum(self._rwork[0][:size_w])

        # --- Reduce energy and enstrophy values overs all proc ---
        # two ways : numpy or classical. Todo : check perf and comm
        sendbuff = npw.zeros((2))
        recvbuff = npw.zeros((2))
        sendbuff[:] = [local_energy, local_enstrophy]
        #
        self.velocity.topology.comm.Allreduce(sendbuff, recvbuff)
        # the other way :
        #energy = self.velocity.topology.allreduce(local_energy,
        #                                          hysop.core.mpi_REAL,
        #                                          op=MPI.SUM)
        #enstrophy = self.velocity.topology.allreduce(local_enstrophy,
        #                                             hysop.core.mpi_REAL,
        #                                             op=MPI.SUM)

        # Update global values
        self.energy = recvbuff[0] * self.coeff['Energy']
        self.enstrophy = recvbuff[1] * self.coeff['Enstrophy']

        # Print results, if required
        ite = simulation.current_iteration
        if self._writer is not None and self._writer.do_write(ite):
            self._writer.buffer[0, 0] = simulation.time
            self._writer.buffer[0, 1] = self.energy
            self._writer.buffer[0, 2] = self.enstrophy
            self._writer.write()
