# -*- coding: utf-8 -*-
"""Discrete operator for Poisson problem (fftw based)

.. currentmodule hysop.operator.discrete

"""
from hysop.tools.numpywrappers import npw

from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.operator.discrete.reprojection import Reprojection
from hysop.constants import debug
from hysop.tools.profiler import profile
try:
    from hysop.f2hysop import fftw2py
except ImportError:
    msg = 'fftw package not available for your hysop install.'
    msg += 'Try to recompile with WITH_FFTW=ON'
    raise ImportError(msg)


class PoissonFFT(DiscreteOperator):
    """Discretized Poisson operator based on FFTW.
    See details in hysop.operator.poisson
    """

    @debug
    def __init__(self, output_field, input_field, projection=None,
                 filter_size=None, correction=None, formulation=None, **kwds):
        """Poisson operator, incompressible flow.

        Parameters
        ----------
        output_field : :class:`~hysop.fields.discrete_field.DiscreteField
            solution field
        input_field  : :class:`~hysop.fields.discrete_field.DiscreteField
            right-hand side
        projection : double or tuple, optional
             projection criterion, see notes below.
        filter_size :
        correction : :class:`~velocity_correction.VelocityCorrection_D
        operator used to shift output_field according
          to a given input (fixed) flowrate.
          See hysop.operator.velocity_correction.
        formulation :
        kwds : base class parameters.
        """
        # Base class initialisation
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(PoissonFFT, self).__init__(variables=[output_field, input_field],
                                         **kwds)
        # Solution field
        self.output_field = output_field
        # RHS field
        self.input_field = input_field
        # Solenoidal projection of input_field ?
        self.projection = projection
        # Filter size array = domainLength/(CoarseRes-1)
        self.filter_size = filter_size
        # If 2D problem, input_field must be a scalar
        self._dim = self.output_field.domain.dim
        if self._dim == 2:
            assert self.input_field.nb_components == 1
        assert (self._dim >= 2),\
            "Wrong problem dimension: only 2D and 3D cases are implemented."
        self.correction = correction
        self.formulation = formulation
        self.input = [self.input_field]
        self.output = [self.output_field]

        # The function called during apply
        self.solve = None
        # a sub function ...
        self._solve = None
        self.do_projection = None
        self._select_solve()

    def _select_solve(self):
        """
        TODO : add pressure solver selection
        f(output.nbComponents) + pb type 'pressure-poisson'
        """
        # Multiresolution?
        multires = \
            self.output_field.topology.mesh != self.input_field.topology.mesh
        # connexion to the required apply function
        if self._dim == 2:
            self._solve = self._solve_2d
        elif self._dim == 3:
            # If there is a projection, input_field is also an output
            if self.projection is not None:
                self.output.append(self.input_field)
                if multires:
                    self._solve = self._solve_3d_proj_multires
                else:
                    self._solve = self._solve_3d_proj

                if isinstance(self.projection, Reprojection):
                    self.do_projection = self.do_projection_with_op
                else:
                    self.do_projection = self.do_projection_no_op

            else:
                if multires:
                    self._solve = self._solve_3d_multires
                elif self.formulation is not None:
                    self._solve = self._solve_3d_scalar_fd
                else:
                    self._solve = self._solve_3d

        # swtich to the proper solving method (with or without correction)
        if self.correction is not None:
            self.solve = self._solve_and_correct
        else:
            self.solve = self._solve

    def do_projection_with_op(self, simu):
        """Compute projection criterion and return
        true if projection is needed.
        """
        self.projection.apply(simu)
        ite = simu.current_iteration
        return self.projection.do_projection(ite)

    def do_projection_no_op(self, simu):
        """return true if projection is needed.
        """
        ite = simu.current_iteration
        return ite % self.projection == 0

    def _solve_2d(self, simu=None):
        """ Solve 2D poisson problem, no projection, no correction.
        """
        ghosts_v = self.output_field.topology.ghosts()
        ghosts_w = self.input_field.topology.ghosts()
        self.output_field.data[0], self.output_field.data[1] =\
            fftw2py.solve_poisson_2d(self.input_field.data[0],
                                     self.output_field.data[0],
                                     self.output_field.data[1],
                                     ghosts_w, ghosts_v)

    def _project(self):
        """apply projection onto input_field
        """
        ghosts_w = self.input_field.topology.ghosts()
        self.input_field.data[0], self.input_field.data[1], \
            self.input_field.data[2] = \
            fftw2py.projection_om_3d(self.input_field.data[0],
                                     self.input_field.data[1],
                                     self.input_field.data[2], ghosts_w)

    def _solve_3d_multires(self, simu=None):
        """3D, multiresolution
        """
        # Projects input_field values from fine to coarse grid
        # in frequencies space by nullifying the smallest modes
        vortFilter = npw.copy(self.input_field.data)
        vortFilter[0], vortFilter[1], vortFilter[2] = \
            fftw2py.multires_om_3d(self.filter_size[0], self.filter_size[1],
                                   self.filter_size[2],
                                   self.input_field.data[0],
                                   self.input_field.data[1],
                                   self.input_field.data[2])

        # Solves Poisson equation using filter input_field
        ghosts_v = self.output_field.topology.ghosts()
        ghosts_w = self.input_field.topology.ghosts()
        self.output_field.data[0], self.output_field.data[1],\
            self.output_field.data[2] = \
            fftw2py.solve_poisson_3d(vortFilter[0], vortFilter[1],
                                     vortFilter[2], self.output_field.data[0],
                                     self.output_field.data[1],
                                     self.output_field.data[2],
                                     ghosts_w, ghosts_v)

    def _solve_3d_proj_multires(self, simu):
        """3D, multiresolution, with projection
        """
        if self.do_projection(simu):
            self._project()
        self._solve_3d_multires()

    def _solve_3d_proj(self, simu):
        """3D, with projection
        """
        if self.do_projection(simu):
            self._project()
        self._solve_3d()

    def _solve_3d(self,simu=None):
        """Basic solve
        """
        # Solves Poisson equation using usual input_field
        ghosts_v = self.output_field.topology.ghosts()
        ghosts_w = self.input_field.topology.ghosts()
        self.output_field.data[0], self.output_field.data[1],\
            self.output_field.data[2] =\
            fftw2py.solve_poisson_3d(self.input_field.data[0],
                                     self.input_field.data[1],
                                     self.input_field.data[2],
                                     self.output_field.data[0],
                                     self.output_field.data[1],
                                     self.output_field.data[2],
                                     ghosts_w, ghosts_v)

    def _solve_and_correct(self, simu):
        """Solve Poisson problem and apply correction on velocity.
        """
        self._solve(simu.current_iteration)
        self.correction.apply(simu)

    def _solve_3d_scalar_fd(self, simu=None):
        """solve poisson-pressure like problem
        input = 3D vector field
        output = 3D scalar field
        """
        # Compute rhs = f(input) inplace
        # --> output == rhs
        # Call fftw filter
        # !!! pressure3d use the same arg for input and output
        # ---> input_field will be overwritten
        ghosts = self.output_field.topology.ghosts()
        self.output_field.data[0] = fftw2py.pressure_3d(
            self.input_field.data[0], ghosts)

    def _solve_3d_scalar(self, simu=None):
        """solve poisson-pressure like problem
        input = 3D vector field
        output = 3D scalar field
        """
        # # Call fftw filter
        # self._output_field.data[0] = fftw2py.solve_poisson_3d_pressure(
        #     self._input_field.data[0],
        #     self._input_field.data[1],
        #     self._input_field.data[2])
        pass

    @debug
    @profile
    def apply(self, simulation=None):
        self.solve(simulation)

    def finalize(self):
        """
        Clean memory (fftw plans and so on)
        """
        pass
        #fftw2py.clean_fftw_solver(self.output_field.dimension)
