# -*- coding: utf-8 -*-
"""Discrete Spectrum operator using FFTW
"""
from hysop.operator.discrete.discrete import DiscreteOperator
from hysop.constants import debug, np, hysop.core.mpi_REAL
from hysop.tools.profiler import profile
from hysop.tools.numpywrappers import npw
from hysop.core.mpi import MPI

try:
    from hysop.f2hysop import fftw2py
except ImportError:
    msgE = 'fftw package not available for your hysop install.'
    msgE += 'Try to recompile with WITH_FFTW=ON'
    raise ImportError(msgE)


class FFTSpectrum(DiscreteOperator):
    """Discretized Spectrum operator based on FFTW.
    """
    @debug
    def __init__(self, field, **kwds):
        """

        Parameters
        ----------
        field: :class:`~hysop.fields.discrete_field.DiscreteField`
            the input field for which spectrum will be computed
        """
        # Discretization of the input field
        self.field = field
        msg = 'Spectrum error: implemented only for 3D problems.'
        assert self.field.dimension == 3, msg
        if self.field.nb_components > 1:
            raise AttributeError("Vector case not yet implemented.")
        # Base class initialisation
        assert 'variables' not in kwds, 'variables parameter is useless.'
        super(FFTSpectrum, self).__init__(variables=[field],
                                          **kwds)
        self.input = [self.field]
        l = np.min(self.field.topology.mesh.discretization.resolution)
        self._tmp = npw.zeros(((l - 1) / 2, ))
        self._kx = npw.zeros(((l - 1) / 2, ))
        self.res = npw.zeros(((l - 1) / 2, ))

    @debug
    @profile
    def apply(self, simulation=None):
        assert simulation is not None, \
            "Missing simulation parameter."
        ite = simulation.current_iteration
        ghosts = self.field.topology.ghosts()
        fftw2py.spectrum_3d(self.field.data[0],
                            self._tmp, self._kx,
                            ghosts, np.min(self.domain.length))
        if self.field.topology.size == 1:
            self.res[...] = self._tmp
        else:
            self.field.topology.comm.Reduce(
                [self._tmp, self.res.shape[0], hysop.core.mpi_REAL],
                [self.res, self.res.shape[0], hysop.core.mpi_REAL],
                op=MPI.SUM, root=0)

        if self._writer is not None and self._writer.do_write(ite):
            nbc = self.res.shape[0]
            self._writer.buffer[0, 0:nbc] = self._kx
            self._writer.buffer[0, nbc:] = self.res
            self._writer.write()

    def finalize(self):
        """Clean memory (fftw plans and so on)
        """
        pass
