"""Tests for non-gpu Advection"""

from hysop.tools.parameters import Discretization
from hysop.domain.box import Box
from hysop.fields.continuous_field import Field
from hysop.problem.simulation import Simulation
from hysop.operator.advection import Advection
from hysop.tools.numpywrappers import npw
import numpy as np


def find_min_work(velo, fields, topo):
    """Get minimal space required for
    internal buffers in advection
    """
    op = Advection(velocity=velo,
                   advected_fields=fields, discretization=topo)
    op.discretize()
    iwork = []
    rwork = []
    work_prop = op.get_work_properties()
    if work_prop['rwork'] is not None:
        for sh in work_prop['rwork']:
            incr_sh = np.asarray(sh)
            incr_sh[0] += 143
            rwork.append(npw.zeros(incr_sh))

    if work_prop['iwork'] is not None:
        for sh in work_prop['iwork']:
            incr_sh = np.asarray(sh)
            incr_sh[0] += 1343
            iwork.append(npw.int_zeros(sh))
        return iwork, rwork

g = 2
d2d = Discretization([17, 17], [g, g])
d3d = Discretization([17, 33, 29], [g, g, g])
d2d_2 = Discretization([25, 17], [2, 2])
d3d_2 = Discretization([25, 33, 29], [2, 2, 2])
box_2d = Box(length=[2., 2.], origin=[-1., -1.])
box_3d = Box(length=[2., 4., 1.], origin=[-1., -2., 0.])
topo_2d = box_2d.create_topology(d2d, cutdir=[False, True])
topo_3d = box_3d.create_topology(d3d, cutdir=[False, False, True])
topo_2d_2 = box_2d.create_topology(d2d_2, cutdir=[False, True])
topo_3d_2 = box_3d.create_topology(d3d_2, cutdir=[False, False, True])
velo_2d = Field(domain=box_2d, name='Velocity2D',
                formula=lambda x, y: (0., 0.), is_vector=True)
velo_3d = Field(domain=box_3d, name='Velocity3D',
                formula=lambda x, y: (0., 0., 0.), is_vector=True)
scal_2d = Field(domain=box_2d, name='Scal2D')
scal_3d = Field(domain=box_3d, name='Scal3D')
scal_2d_2 = Field(domain=box_2d, name='Scal2D2')
scal_3d_2 = Field(domain=box_3d, name='Scal3D2')
global_iwork_2d, global_rwork_2d = find_min_work(velo_2d, scal_2d, topo_2d)
global_iwork_3d, global_rwork_3d = find_min_work(velo_3d, scal_3d, topo_3d)
simu = Simulation(start=0., end=0.1, nb_iter=1)


def build_advection(velo, fields, discr, discr_fields=None,
                    rwork=None, iwork=None):
    """declare, discretize and setup advection operator

    Parameters
    ----------

    velo, fields : velocity field and list of advected fields
    discr : chosen discretization for velocity
    discr_fields : chosen discretization for advected_fields
       if None == discr
    rwork, iwork : pre-allocated work arrays
       if None, let operator do the allocation job.

    """
    # Build and discretize advection
    op = Advection(velocity=velo,
                   advected_fields=fields, discretization=discr,
                   discretization_fields=discr_fields)
    op.discretize()
    # checks
    assert velo in op.input
    assert velo not in op.output
    if isinstance(fields, list):
        for f in fields:
            assert f in op.input
            assert f in op.output
    else:
        assert fields in op.input
        assert fields in op.output
    assert len(op.discrete_op) == velo.domain.dim
    # iwork = None
    # rwork = None
    # if use_work:
    #     work_prop = op.get_work_properties()
    #     if work_prop['rwork'] is not None:
    #         rwork = []
    #         for sh in work_prop['rwork']:
    #             rwork.append(npw.zeros(sh))

    #     if work_prop['iwork'] is not None:
    #         rwork = []
    #         for sh in work_prop['iwork']:
    #             iwork.append(npw.zeros(sh))

    #     work_prop = op.get_work_properties()
    op.setup(rwork, iwork)
    dop = op.discrete_op[0]
    rw0 = dop._rwork
    iw0 = dop._iwork
    for i in xrange(1, velo.domain.dim):
        dop = op.discrete_op[i]
        rw = dop._rwork
        iw = dop._iwork
        assert len(rw) == len(rw0)
        assert len(iw) == len(iw0)
        for j in xrange(len(rw)):
            assert npw.arrays_share_data(rw[j], rw0[j])
        for j in xrange(len(iw)):
            assert npw.arrays_share_data(iw[j], iw0[j])

    return op


def test_build_2d_1():
    """single resolution build, 2d
    """
    build_advection(velo_2d, scal_2d, d2d)


def test_build_2d_2():
    """single resolution build, from topo, 2d"""
    build_advection(velo_2d, [scal_2d, scal_2d_2], topo_2d)


def test_build_2d_3():
    """multi resolution build, 2d
    """
    build_advection(velo_2d, [scal_2d, scal_2d_2], d2d_2, d2d)


def test_build_2d_4():
    """multi resolution build, from topo, 2d"""
    build_advection(velo_2d, scal_2d, topo_2d_2, topo_2d)


def test_build_2d_5():
    """multi resolution build, from topo, 2d, with work"""
    op = build_advection(velo_2d, scal_2d, topo_2d_2, topo_2d,
                         rwork=global_rwork_2d, iwork=global_iwork_2d)
    for i in xrange(velo_2d.domain.dim):
        dop = op.discrete_op[i]
        rw = dop._rwork
        iw = dop._iwork
        for j in xrange(len(rw)):
            assert npw.arrays_share_data(rw[j], global_rwork_2d[j])
        for j in xrange(len(iw)):
            assert npw.arrays_share_data(iw[j], global_iwork_2d[j])


def test_build_3d_1():
    """single resolution build, 3d
    """
    build_advection(velo_3d, scal_3d, d3d)


def test_build_3d_2():
    """single resolution build, from topo, 3d"""
    build_advection(velo_3d, [scal_3d, scal_3d_2], topo_3d)


def test_build_3d_3():
    """multi resolution build, 3d
    """
    build_advection(velo_3d, [scal_3d, scal_3d_2], d3d_2, d3d)


def test_build_3d_4():
    """multi resolution build, from topo, 3d"""
    build_advection(velo_3d, scal_3d, topo_3d_2, topo_3d)


def test_build_3d_5():
    """multi resolution build, from topo, 3d, with work"""
    op = build_advection(velo_3d, scal_3d, topo_3d_2, topo_3d,
                         rwork=global_rwork_3d, iwork=global_iwork_3d)
    for i in xrange(velo_3d.domain.dim):
        dop = op.discrete_op[i]
        rw = dop._rwork
        iw = dop._iwork
        for j in xrange(len(rw)):
            assert npw.arrays_share_data(rw[j], global_rwork_3d[j])
        for j in xrange(len(iw)):
            assert npw.arrays_share_data(iw[j], global_iwork_3d[j])


def test_apply_2d():
    """advection of a field ... at null velocity, 2d.
    """
    op = build_advection(velo_2d, scal_2d, d2d)
    op.apply(simu)
