"""Description of a problem, i.e. a sequence of operators
"""
from hysop.constants import debug
import cPickle
from hysop import __VERBOSE__, __GPU_ENABLED__
from hysop.operator.redistribute import Redistribute, RedistributeIntra
from hysop.tools.profiler import profile, Profiler
from hysop.core.mpi import main_rank
if __GPU_ENABLED__:
    from hysop.backend.device.opencl.gpu_transfer import DataTransfer
else:
    DataTransfer = int


class Problem(object):
    """Problem representation.

    A problem is a sequence of operators applied on a set of
    continuous fields.
    """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, operators, simulation,
                 dump_freq=100, name=None):
        """

        Parameters
        ----------
        operators : list of :class:`~hysop.operator.continuous.Continuous`
            sequence of operators used to define the problem
        simulation : :class:`~hysop.problem.simulation.Simulation`
            description of the time discretisation
        dump_freq : real; optional
            frequency of dump to file for the problem.
            Set dump_freq = -1 for no dumps. Default = 100,
            i.e. every 100 time steps.
        name : string, optional
            id for the problem
        """
        # Problem name
        self.name = name
        # Problem operators
        self.operators = operators
        # Computes time step and manage iterations
        self.simulation = simulation

        # all variables must be defined on the same domain.
        vref = self.operators[0].variables.keys()[0]
        self.domain = vref.domain
        msg = "All variables must be defined on the same domain."
        for op in self.operators:
            for v in (v for v in op.variables if v is not vref):
                print id(v.domain), id(self.domain)
                print v.domain, self.domain
                if self.domain is not v.domain:
                    raise ValueError(msg)
        # A list of variables that must be initialized before
        # any call to op.apply()
        self.input = []
        # call to problem.dump frequency during apply.
        if dump_freq >= 0:
            # dump problem every self.dump_freq iter
            self.dump_freq = dump_freq
        else:
            self.dump_freq = None

        # Id for the problem. Used for dump file name.
        if name is None:
            self.name = 'HySoPPb'
        else:
            self.name = name
        # Object to store computational times of lower level functions
        self.profiler = Profiler(self, self.domain.task_comm)
        # Default file name prefix for dump.
        self.filename = str(self.name)
        self._filedump = self.filename + '_rk_' + str(main_rank)

        # Flag : true when operators for computation are up
        # and when variables are initialized (i.e. after a call to pre_setup)
        # Note : 3 categories of op : computation (stretching, poisson ...),
        # and data distribution (Redistribute)
        self._is_ready = False

    @debug
    @profile
    def setup(self):
        """Prepare operators (create topologies, allocate memories ...)
        """
        # Set up for 'computational' operators
        if not self._is_ready:
            self.pre_setup()
        print "Fin setup op"
        # for v in self.input:
        #     v.initialize()

        # other operators
        for op in self.operators:
            if isinstance(op, RedistributeIntra) or \
               isinstance(op, DataTransfer):
                op.setup()

        for op in self.operators:
            if isinstance(op, Redistribute):
                op.setup()

        if __VERBOSE__ and main_rank == 0:
            print ("====")

    def pre_setup(self):
        """Discretization and setup for computational operators
         and fields initialization
        """
        if self._is_ready:
            return

        for op in self.operators:
            if not isinstance(op, Redistribute) and \
               not isinstance(op, DataTransfer):
                op.discretize()
                op.setup()

        if __VERBOSE__ and main_rank == 0:
            print ("==== Variables initialization ====")

        # Build the list of 'input' variables, that must be initialized.
        # 'input' fields are variables that are not output of
        # any previous operators in the operator sequence.
        self.input = []

        # First, all operators input vars are appended to input list
        for op in self.operators:
            for v in op.input:
                if v not in self.input:
                    self.input.append(v)

        # Then starting from the last op, vars which are 'input' of
        # an operator AND output of a previous op are removed
        for op in self.operators[-1::-1]:
            for v in self.input:
                if v in op.output:
                    self.input.remove(v)
                if v in op.input:
                    self.input.append(v)

        self._is_ready = True

    @debug
    @profile
    def solve(self):
        """Apply all operators of the problem, for all simulation
        time steps

        * init simulation
        * for each time step:
            * apply all operators
            * increment time step
            * dump problem (for concerned values, depend on dump_freq)

        """
        # Init simulation
        self.simulation.initialize()
        if main_rank == 0:
            print ("\n\n Start solving ...")
        # Run simulation
        while not self.simulation.is_over:
            if main_rank == 0:
                self.simulation.print_state()

            for op in self.operators:
                if __VERBOSE__:
                    print (main_rank, op.name)
                op.apply(self.simulation)
            testdump = False
            if self.dump_freq is not None:
                testdump = \
                    self.simulation.current_iteration % self.dump_freq is 0
            self.simulation.advance()
            if testdump:
                self.dump()

    @debug
    def finalize(self):
        """
        Finalize method
        """
        if main_rank == 0:
            print ("\n\n==== End ====")
        for op in self.operators:
            op.finalize()

        var = []
        for op in self.operators:
            for v in op.variables:
                if not v in var:
                    var.append(v)
        for v in var:
            v.finalize()
        self.profiler.summarize()
        if main_rank == 0:
            print ("===\n")

    def get_profiling_info(self):
        for op in self.operators:
            self.profiler += op.profiler
        for op in self.operators:
            for v in op.variables:
                self.profiler += v.profiler

    def __str__(self):
        """ToString method"""
        s = "Problem based on\n"
        s += str(self.domain.topologies)
        s += "with following operators : \n"
        for op in self.operators:
            s += str(op)
        return s

    def dump(self, filename=None):
        """Serialize some data of the problem to file
        (only data required for a proper restart, namely fields in self.input
        and simulation).

        Parameters
        ----------
        filename : string
            prefix for output file. Real name = filename_rk_N,
        N being current process number. If None use default value from problem
        parameters (self.filename)
        """
        if filename is not None:
            self.filename = filename
            self._filedump = filename + '_rk_' + str(main_rank)
        db = open(self._filedump, 'wb')
        cPickle.dump(self.simulation, db)
        # TODO : review dump process using hdf files instead of pickle.
        # for v in self.input:
        #     v.hdf_dump(self.filename)

    def restart(self, filename=None):
        """
        Load serialized data to restart from a previous state.
        self.input variables and simulation are loaded.
        @param  filename : prefix for downloaded file.
        Real name = filename_rk_N, N being current process number.
        If None use default value from problem
        parameters (self.filename)
        """
        if filename is not None:
            self.filename = filename
            self._filedump = filename + '_rk_' + str(main_rank)
        db = open(self._filedump, 'r')
        self.simulation = cPickle.load(db)
        self.simulation.reset()
        for v in self.input:
            print ("load ...", self.filename)
            v.load(self.filename)

        for op in self.operators:
            if isinstance(op, Redistribute):
                op.setup()

    def setDumpFreq(self, freq):
        """set rate of problem.dump call (every 'rate' iteration)
        :param : the frequency of output. <0 or None means no output.
        """
        self.dump_freq = freq
        if freq < 0:
            self.dump_freq = None
