"""Tests simulation time loop parameters
"""
from hysop.problem.simulation import Simulation, eps
from hysop.tools.io_utils import Writer, IOParams, IO
from hysop.core.mpi import main_rank
import numpy as np


def run_simu(s):
    i = 0
    tref = s.start
    while not s.is_over:
        assert s.current_iteration == i
        assert s.time == tref + s.time_step
        tref = s.time
        i += 1
        s.advance()


def test_simu_default():
    s = Simulation(nb_iter=10)
    assert s.time == 0.
    assert s.end == 1.
    assert np.allclose(s.time_step, 1. / 10., rtol=eps)
    s.initialize()
    assert np.allclose(s.time, s.time_step)

    run_simu(s)

    assert np.allclose(s.time, s.end)
    assert s.current_iteration == 9
    assert s.nb_iter == 10

    s.finalize()
    assert np.allclose(s.time, s.end)
    assert s.current_iteration == -1
    assert s.is_over

    s.initialize()
    assert s.current_iteration == 0
    assert not s.is_over
    assert np.allclose(s.time, 0.1)
    assert s.time == 0.1


def test_simu_2():
    s = Simulation(time_step=0.1, max_iter=5)
    assert s.time == 0.
    assert s.end == 1.
    assert np.allclose(s.time_step, 1. / 10.)
    s.initialize()
    assert np.allclose(s.time, s.time_step)

    run_simu(s)

    assert s.current_iteration == 4
    assert np.allclose(s.time, s.start + 5 * s.time_step)


def test_simu_adapt():
    s = Simulation(time_step=0.1)
    s.initialize()
    i = 0
    tref = s.start
    while not s.is_over:
        assert s.current_iteration == i
        assert s.time == tref + s.time_step
        tref = s.time
        if s.current_iteration == 5:
            s.update_time_step(s.time_step * 0.5)
        s.advance()
        i += 1

    assert np.allclose(s.time, s.end)
    assert np.allclose(s.time_step, 0.05)
    assert s.current_iteration == 13


def test_simu_incr():
    simu = Simulation(start=0.0, end=1.0, nb_iter=10)
    io_params = IOParams(filename='temp_test', frequency=2,
                         fileformat=IO.ASCII)
    wr = Writer(io_params)
    leader = io_params.io_leader
    if main_rank == leader:
        assert wr.do_write(simu.current_iteration)
    else:
        assert not wr.do_write(simu.current_iteration)

    simu.initialize()

    assert not wr.do_write(simu.current_iteration)

    count = 1
    while not simu.is_over:
        if count % 2 == 0 and main_rank == leader:
            assert wr.do_write(simu.current_iteration)
        else:
            assert not wr.do_write(simu.current_iteration)
        simu.advance()
        count += 1
    assert simu.current_iteration == 9
    simu.finalize()

    if main_rank == leader:
        assert wr.do_write(simu.current_iteration)
    else:
        assert not wr.do_write(simu.current_iteration)


def test_simu_incr2():
    simu = Simulation(start=0.0, end=1.0, nb_iter=10)
    io_params = IOParams(filename='temp_test', frequency=3,
                         fileformat=IO.ASCII)
    leader = io_params.io_leader
    wr = Writer(io_params)
    if main_rank == leader:
        assert wr.do_write(simu.current_iteration)
    else:
        assert not wr.do_write(simu.current_iteration)
    simu.update_time_step(0.10000000001)
    simu.initialize()

    assert not wr.do_write(simu.current_iteration)

    count = 1
    while not simu.is_over:
        if count % 3 == 0 and main_rank == leader:
            assert wr.do_write(simu.current_iteration)
        else:
            assert not wr.do_write(simu.current_iteration)
        simu.advance()
        count += 1
    assert simu.current_iteration == 9
    simu.finalize()
    if main_rank == leader:
        assert wr.do_write(simu.current_iteration)
    else:
        assert not wr.do_write(simu.current_iteration)
